from collections import OrderedDict


class iecGroup(OrderedDict):
    def __init__(self, *arg, **kw):
        super(iecGroup, self).__init__(*arg, **kw)
        if len(arg) > 0 and isinstance(arg[0], iecGroup):
            for key, value in arg[0].items():
                # super(iecGroup, self).__init__(key) = value
                self[key] = value

    def copy(self, fromObject):
        for key, value in fromObject.items():
            # super(iecGroup, self).__init__(key) = value
            self[key] = value

    def outputString(self):
        outputString = ""
        for key, value in self.items():
            outputString += value.outputString()
        return outputString



import os
import shutil

from pynumeca.constants import constants


def a5_mesh_from_geomturbo(
    source_trb_path: str,
    geomturbo_path: str,
    # output_trb_path: str,
    output_igg_path: str,
    igg_version: str = constants.version,
    cores: int = 1,
):
    # shutil.copy(source_trb_path, output_trb_path)
    cmd = (
        f"igg{igg_version}"
        + " -print -real-batch -autogrid5"
        + " -numproc "
        + str(cores)
        + " -trb "
        + '"'
        + source_trb_path
        + '"'
        + " -geomTurbo "
        + '"'
        + geomturbo_path
        + '"'
        + " -mesh "
        + '"'
        + output_igg_path
        + '"'
    )

    print("Running command '" + cmd + "'")
    os.system(cmd)



import os
import shutil

import tqdm


def collect_design_3d_results(root_path: str, target_path: str, reset: bool = False):
    if reset:
        if os.path.exists(target_path):
            shutil.rmtree(target_path)

    if not os.path.exists(target_path):
        os.mkdir(target_path)

    target_geoms_path = os.path.join(target_path, "geometries")
    os.mkdir(target_geoms_path)

    target_performance_path = os.path.join(target_path, "performance")
    os.mkdir(target_performance_path)

    flow_files = []
    for file in os.listdir(root_path):
        if "flow" in file or "design" in file:
            if os.path.isdir(os.path.join(root_path, file)):
                flow_files.append(os.path.join(root_path, file))

    if len(flow_files) == 0:
        return ValueError("No flow/design files found")

    for file in tqdm.tqdm(flow_files):
        mesh_path = os.path.join(file, "_mesh")

        if not os.path.exists(mesh_path):
            continue

        for item in os.listdir(mesh_path):
            if "geomturbo" in item.lower():
                geom_path = os.path.join(mesh_path, item)
                break

        shutil.copy(
            src=geom_path,
            dst=os.path.join(
                target_geoms_path,
                os.path.basename(file)
                + os.path.splitext(os.path.basename(geom_path))[1],
            ),
        )

        perf_path = os.path.join(target_performance_path, os.path.basename(file))
        os.mkdir(perf_path)

        simulations_path = []
        for f in os.listdir(file):
            if os.path.isdir(os.path.join(file, f)):
                _path = os.path.join(file, f)
                _path_content = os.listdir(_path)
                for c in _path_content:
                    if (
                        os.path.splitext(c)[1] == ".res"
                        or os.path.splitext(c)[1] == ".mf"
                    ):
                        simulations_path.append(_path)
                        break

        for sim_path in simulations_path:
            for f in os.listdir(sim_path):
                if os.path.splitext(f)[1] == ".res" or os.path.splitext(f)[1] == ".mf":
                    shutil.copy(
                        src=os.path.join(sim_path, f), dst=os.path.join(perf_path, f)
                    )



import pandas as pd


def find_between(s, first, last):
    try:
        start = s.index(first) + len(first)
        end = s.index(last, start)
        return s[start:end]
    except ValueError:
        return ""


def plan_to_dataframe(plan_path: str, log_head: bool = False):
    """
    :param plan_path: .plan file path
    :param log_head: logging or not the dataframe head
    :return: pandas dataframe representing the specified .plan file

    Converting a Numeca .plan file into a pandas dataframe.plan

    """

    # Reading the plan file
    with open(plan_path, "r") as f:
        content = f.read()

    # fixing .plan variable spacing
    for i in range(1, 10):
        content = content.replace(i * " ", " ")

    content = content.replace("\t", " ")

    # Reading dataframe columns
    edp = (
        find_between(content, "# Design parameters", "# Responses")
        .replace("\n", "")
        .split()
    )
    responses = (
        find_between(content, "# Responses", "# Experiments ").replace("\n", "").split()
    )
    info = (
        find_between(content, "# Experiments (", ")")
        .replace("\n", "")
        .replace(",", "")
        .split()
    )

    # Reading raw dataframe
    df_raw = [
        [item for item in line.split(" ") if item != ""]
        for line in content[content.index(")") + 2 :].split("\n")
    ]

    # Merging columns
    cols = info[:2]
    cols.extend(edp)
    cols.extend(responses)
    cols.append(info[-1])

    # Final DataFrame
    df = pd.DataFrame(df_raw, columns=cols)
    print(f"DataFrame shape: ", {df.shape})
    if log_head:
        print(df.head())

    return df



import copy

from pynumeca.reader.iecGroup import iecGroup
from pynumeca.reader.sectionEntry import sectionEntry


class niBladeGeometryEntry(iecGroup):
    def __init__(self, *args):
        super(niBladeGeometryEntry, self).__init__()
        self.numberOfSectionsInteger = 0
        self.suctionHeader = None
        self.sectionalHeader = None
        self.numberOfSectionsEntry = None
        self.headerList = []
        self.suctionList = []
        self.pressureList = []
        self.footer = None

        if len(args) > 0 and isinstance(
            args[0], iecGroup
        ):  # Deve essere un gruppo NiBladeGeometry
            entryIndex = 0
            totalLength = len(list(args[0].items()))
            while True:
                if entryIndex < totalLength:
                    key, value = list(args[0].items())[entryIndex]
                else:
                    break
                if "suction" in key:
                    self.suctionHeader = value
                    self.sectionalHeader = list(args[0].items())[entryIndex + 1][1]
                    self.numberOfSectionsEntry = list(args[0].items())[entryIndex + 2][
                        1
                    ]
                    self.numberOfSectionsInteger = int(self.numberOfSectionsEntry.key)

                    for item in range(self.numberOfSectionsInteger):
                        previousPoints = 0
                        for subItem in range(item):
                            previousPoints += (
                                self.suctionList[subItem].numberOfPointsInt + 3
                            )
                        section = self.extractSection(
                            args[0], entryIndex + 3 + previousPoints
                        )

                        self.suctionList.append(section)

                    pressureRowIndex = (
                        entryIndex
                        + previousPoints
                        + self.suctionList[-1].numberOfPointsInt
                        + 6
                    )

                    self.pressureHeader = list(args[0].items())[pressureRowIndex][1]

                    for item in range(self.numberOfSectionsInteger):
                        previousPoints = 0
                        for subItem in range(item):
                            previousPoints += (
                                self.pressureList[subItem].numberOfPointsInt + 3
                            )
                        section = self.extractSection(
                            args[0], pressureRowIndex + previousPoints + 3
                        )
                        self.pressureList.append(section)
                    entryIndex = (
                        pressureRowIndex
                        + 6
                        + previousPoints
                        + self.pressureList[-1].numberOfPointsInt
                    )
                    self.footer = list(args[0].items())[entryIndex][1]
                    break

                else:
                    self.headerList.append(value)

                entryIndex += 1

    def extractSection(self, entry, entryIndex):
        section = sectionEntry()
        section.headerComment = list(entry.items())[entryIndex][1]
        section.referenceFrame = list(entry.items())[entryIndex + 1][1]
        section.numberOfPointsEntry = list(entry.items())[entryIndex + 2][1]
        section.numberOfPointsInt = int(section.numberOfPointsEntry.key)
        section.leadingSpaceString = list(entry.items())[entryIndex + 3][
            1
        ].leadingSpaceString
        section.trailingSpaceString = list(entry.items())[entryIndex + 3][
            1
        ].trailingSpaceString
        section.X2YSpaceString = list(entry.items())[entryIndex + 3][
            1
        ].tagValueSpaceString[0]
        section.Y2ZSpaceString = list(entry.items())[entryIndex + 3][
            1
        ].tagValueSpaceString[1]
        for row in range(section.numberOfPointsInt):
            section.X.append(float(list(entry.items())[entryIndex + 3 + row][1].key))
            section.Y.append(
                float(list(entry.items())[entryIndex + 3 + row][1].value[0])
            )
            section.Z.append(
                float(list(entry.items())[entryIndex + 3 + row][1].value[1])
            )
        return section

    def outputString(self):
        outputString = ""
        for row in self.headerList:
            outputString += row.outputString()
        outputString += self.suctionHeader.outputString()
        outputString += self.sectionalHeader.outputString()
        outputString += self.numberOfSectionsEntry.outputString()
        for section in self.suctionList:
            outputString += section.outputString()
        outputString += self.pressureHeader.outputString()
        outputString += self.sectionalHeader.outputString()
        outputString += self.numberOfSectionsEntry.outputString()
        for section in self.pressureList:
            outputString += section.outputString()
        outputString += self.footer.outputString()

        return outputString

    def setNumberOfSections(self, number):
        diff = self.numberOfSectionsInteger - number
        self.numberOfSectionsInteger = number
        self.numberOfSectionsEntry.key = str(number)
        if diff == 0:
            pass
        elif diff < 0:
            for i in range(abs(diff)):
                self.suctionList.append(copy.deepcopy(self.suctionList[-1]))
                self.pressureList.append(copy.deepcopy(self.pressureList[-1]))
        elif diff > 0:
            del self.suctionList[(-1) * diff :]
            del self.pressureList[(-1) * diff :]



import numpy as np

from pynumeca.reader.iecEntry import iecEntry


class sectionEntry(iecEntry):
    def __init__(self, *args):
        super(sectionEntry, self).__init__()
        self.headerComment = None
        self.numberOfPointsInt = 0
        self.numberOfPointsEntry = iecEntry()
        self.referenceFrame = "XYZ"
        self.leadingSpaceString = ""
        self.X2YSpaceString = ""
        self.Y2ZSpaceString = ""
        self.trailingSpaceString = ""
        self.X = []
        self.Y = []
        self.Z = []
        self.R = []
        self.THETA = []

    def __repr__(self):
        pass

    def __str__(self):
        pass

    def outputString(self):
        outputString = ""
        outputString += self.headerComment.outputString()
        outputString += self.referenceFrame.outputString()
        outputString += self.numberOfPointsEntry.outputString()
        for i in range(self.numberOfPointsInt):
            outputString += (
                self.leadingSpaceString
                + str(self.X[i])
                + self.X2YSpaceString
                + str(self.Y[i])
                + self.Y2ZSpaceString
                + str(self.Z[i])
                + self.trailingSpaceString
            )
        return outputString

    def updateArrays(self, newX, newY, newZ):
        if len(newX) != len(newY) or len(newY) != len(newZ):
            print("ERRORE")
            exit()
        self.numberOfPointsInt = len(newX)
        self.numberOfPointsEntry.key = str(self.numberOfPointsInt)
        self.X = newX
        self.Y = newY
        self.Z = newZ

    def updateArraysCyl(self, newR, newTHETA, newZ):
        if len(newR) != len(newTHETA) or len(newTHETA) != len(newZ):
            print("ERRORE")
            exit()
        self.numberOfPointsInt = len(newR)
        self.numberOfPointsEntry.key = str(self.numberOfPointsInt)
        self.R = newR
        self.THETA = newTHETA
        self.Z = newZ
        for index in range(len(newR)):
            self.X[index] = self.R[index] * np.cos(self.THETA[index])
            self.Y[index] = self.R[index] * np.sin(self.THETA[index])



from pynumeca.reader import (
    iecEntry,
    iecGroup,
    niBladeGeometryEntry,
    numecaEntry,
    numecaParser,
    sectionEntry,
    zrCurveEntry,
)

__all__ = [
    "iecEntry",
    "iecGroup",
    "niBladeGeometryEntry",
    "numecaEntry",
    "numecaParser",
    "sectionEntry",
    "zrCurveEntry",
]



from typing import Optional

import numpy as np
import pyfluids
from pyfluids import Fluid, FluidsList, Input


class Boundaries(object):
    """
    Class representing the boundaries of a turbomachine.
    """

    update_enabled = False
    __R = 8.314462618

    def __init__(
        self,
        m: float,
        pt_in: float,
        tt_in: float,
        cp: Optional[float] = None,
        mu: Optional[float] = None,
        R: Optional[float] = None,
        fluid: Optional[pyfluids.fluids.fluid.FluidsList] = None,
    ):
        """
        Initialize the boundaries of a turbomachine.

        Parameters:
            - fluid (pyfluids.fluids.fluid.Fluid): fluid type
            - m (float): Mass flow rate
            - pt_in (float): Total pressure at the inlet of the turbomachine [Pa]
            - tt_in (float): Total temperature at the inlet of the turbomachine [K]
            - R (float): Specific gas constant
            - k (float): Specific heat ratio
            - cp (float): Specific heat at constant pressure
            - mu (float): Dynamic viscosity
        """

        self.m = m
        self.pt_in = pt_in
        self.tt_in = tt_in

        self.fluid = fluid

        if fluid is not None:
            actual_fluid = self.get_actual_fluid()

            self.cp = actual_fluid.specific_heat
            self.mu = actual_fluid.dynamic_viscosity
            self.R = self.get_gas_constant(actual_fluid)
        else:
            self.cp = cp
            self.mu = mu
            self.R = R

        self.k = self.cp / (self.cp - self.R)
        self.rho = pt_in / (self.R * tt_in)

        if isinstance(tt_in, np.ndarray):
            self.a = np.sqrt(tt_in * self.R * self.k)
        else:
            self.a = float(np.sqrt(tt_in * self.R * self.k))

        self.update_enabled = True

    def get_gas_constant(self, fluid: pyfluids.Fluid):
        return self.__R / fluid.molar_mass

    def get_actual_fluid(self):
        return Fluid(self.fluid).with_state(
            Input.pressure(self.pt_in),
            # CASTING TO CELSIUS
            Input.temperature(self.tt_in - 273.15),
        )

    def phi(self, omega: float, de: float) -> np.ndarray:
        """
        Compute the flow coefficient.

        Parameters:
            - omega (float): Angular velocity
            - de (float): Diameter of the turbomachine.

        Returns:
            - np.ndarray: The flow coefficient.
        """
        return np.array(self.m / (self.rho * omega * de**3)).reshape(-1, 1)

    def re(self, omega: float, de: float) -> np.ndarray:
        """
        Compute the Reynolds number.

        Parameters:
            - omega (float): Angular velocity
            - de (float): Diameter of the turbomachine.

        Returns:
            - np.ndarray: The Reynolds number.
        """
        return np.array(self.rho * omega * de**2 / self.mu).reshape(-1, 1)

    def ma(self, omega: float, de: float) -> np.ndarray:
        """
        Compute the Mach number.

        Parameters:
            - omega (float): Angular velocity
            - de (float): Diameter of the turbomachine.

        Returns:
            - np.ndarray: The Mach number.
        """
        return np.array(omega * de / self.a * 0.5).reshape(-1, 1)

    def psi_t(self, torque: float, omega: float, de: float, eta: float) -> float:
        """
        Compute the work coefficient.

        Parameters:
            - torque (float): Torque
            - omega (float): Angular velocity
            - de (float): Diameter of the turbomachine.
            - eta (float): Efficiency

        Returns:
            - float: The work coefficient.
        """
        return (torque * eta) / (self.m * omega * de**2)

    def psi_p(self, power: float, omega: float, de: float, eta: float) -> float:
        """
        Compute the work coefficient.

        Parameters:
            - power (float): Power
            - omega (float): Angular velocity
            - de (float): Diameter of the turbomachine.
            - eta (float): Efficiency

        Returns:
            - float: The work coefficient.
        """
        return (power * eta) / (self.m * omega**2 * de**2)

    def psi_b(self, beta: float, omega: float, de: float):
        """
        Compute the work coefficient.

        Parameters:
            - beta (float): Compression ratio
            - omega (float): Angular velocity
            - de (float): Diameter of the turbomachine.

        Returns:
            - float: The work coefficient.
        """
        return (
            self.cp
            * self.tt_in
            / ((de**2) * (omega**2))
            * (beta ** ((self.k - 1) / self.k) - 1)
        )

    def beta(self, psi_is: float, omega: float, de: float) -> float:
        """
        Compute the compression ratio.

        Parameters:
            - psi_is (float): work coefficient
            - omega (float): Angular velocity
            - de (float): Diameter of the turbomachine.

        Returns:
            - float: The compression ratio.
        """
        return (psi_is * (omega**2) * (de**2) / (self.cp * self.tt_in) + 1) ** (
            self.k / (self.k - 1)
        )



import numpy as np


def car2cil(x: np.ndarray):
    # assert x.shape[-1] == 3, f'car2cil: x.shape[-1] should be 3, but its {x.shape[-1]}'
    y = x.copy()
    y[..., 0] = np.sqrt(x[..., 0] ** 2 + x[..., 1] ** 2)
    y[..., 1] = np.arctan2(x[..., 0], x[..., 1])

    return y


def point_curve_closest_points(point, curve_points):
    distances = np.linalg.norm(curve_points - point, axis=1)
    closest_indices = np.argpartition(distances, 2)[:2]
    return closest_indices


def line_parametric_equation(point1, point2):
    x1, y1, z1 = point1
    x2, y2, z2 = point2
    # line equation in 3D space is x = x1 + t(x2 - x1)
    # y = y1 + t(y2 - y1)
    # z = z1 + t(z2 - z1)
    return np.array([x1, y1, z1]), np.array([x2 - x1, y2 - y1, z2 - z1])


def evaluate_parametric_line(t, line_equation):
    point, direction = line_equation
    return point + t * direction


def euclidean_distance(point, points_array):
    point = np.array(point)
    points_array = np.array(points_array)
    # calculate the difference between the point and each point in the array
    diff = points_array - point
    # square the difference
    sq_diff = diff**2
    # sum the squared difference along each axis
    sum_sq_diff = sq_diff.sum(axis=1)
    # take the square root of the sum
    distance = np.sqrt(sum_sq_diff)
    return distance


def point_curve_distance(point, curve_points, n: int = 1000):
    closest_points = curve_points[point_curve_closest_points(point, curve_points)]

    line = line_parametric_equation(closest_points[0], closest_points[1])

    line_points = []
    for t in np.linspace(0, 1, n):
        line_points.append(evaluate_parametric_line(t, line))

    projection = line_points[np.argmin(euclidean_distance(point, line_points))]
    return euclidean_distance(projection, point.reshape(1, -1))


def fitting_metrics(original_curve, fitted_curve):
    distances = []
    for point in fitted_curve:
        distances.append(point_curve_distance(point, original_curve))
    distances = np.array(distances)

    metrics = {}
    metrics["mean"] = distances.mean()
    metrics["std"] = distances.std()
    metrics["max"] = distances.max()
    metrics["min"] = distances.min()
    metrics["max_index"] = distances.argmax()

    return metrics



from pynumeca.reader.iecGroup import iecGroup


class zrCurveEntry(iecGroup):
    def __init__(self, *args):
        super(zrCurveEntry, self).__init__()
        self.curveType = ""
        self.numberOfPoints = 0
        self.Z = []
        self.R = []
        self.leadingSpaceStringArray = ""
        self.trailingSpaceStringArray = "\n"
        self.middleSpaceStringArray = ""
        self.leadingSpaceStringHeader = ""
        self.trailingSpaceStringHeader = "\n"
        self.headerEntry = None
        self.footerEntry = None

        if len(args) > 0 and isinstance(args[0], iecGroup):
            self.headerEntry = list(args[0].items())[0][1]
            self.footerEntry = list(args[0].items())[-1][1]
            self.curveType = list(args[0].items())[1][1]
            self.numberOfPoints = int(list(args[0].items())[2][1].key)
            for i in range(self.numberOfPoints):
                self.Z.append(float(list(args[0].items())[3 + i][1].key))
                self.R.append(float(list(args[0].items())[3 + i][1].value))
            self.leadingSpaceStringArray = list(args[0].items())[-2][
                1
            ].leadingSpaceString
            self.trailingSpaceStringArray = list(args[0].items())[-2][
                1
            ].trailingSpaceString
            self.middleSpaceStringArray = list(args[0].items())[-2][
                1
            ].tagValueSpaceString
            self.leadingSpaceStringHeader = list(args[0].items())[0][
                1
            ].leadingSpaceString
            self.trailingSpaceStringHeader = list(args[0].items())[0][
                1
            ].trailingSpaceString

    def addPoint(self, *args):
        if len(args) == 1 and isinstance(args[0], tuple):
            self.addPointUpdateLists(args[0][0], args[0][1])
        elif (
            len(args) == 2 and isinstance(args[0], float) and isinstance(args[1], float)
        ):
            self.addPointUpdateLists(args[0], args[1])

    def addPointUpdateLists(self, Z, R):
        self.Z.append(Z)
        self.R.append(R)
        self.numberOfPoints += 1
        list(self.items())[2][1].key = str(self.numberOfPoints)

    def removePoint(self, *args):
        self.Z.pop(*args)
        self.R.pop(*args)
        self.numberOfPoints -= 1
        list(self.items())[2][1].key = str(self.numberOfPoints)

    def outputString(self):
        outputString = ""
        outputString += self.headerEntry.outputString()
        outputString += self.curveType.outputString()
        outputString += (
            self.leadingSpaceStringArray
            + str(self.numberOfPoints)
            + self.trailingSpaceStringArray
        )
        for i in range(self.numberOfPoints):
            outputString += (
                self.leadingSpaceStringArray
                + str(self.Z[i])
                + self.middleSpaceStringArray
                + str(self.R[i])
                + self.trailingSpaceStringArray
            )
        outputString += self.footerEntry.outputString()
        return outputString

    def append(self, new_ZR):
        self.numberOfPoints += new_ZR.numberOfPoints - 1
        self.Z.extend(new_ZR.Z[1:])
        self.R.extend(new_ZR.R[1:])

    def updateArrays(self, newZ, newR):
        if len(newZ) != len(newR):
            print("ERRORE")
            exit()
        self.numberOfPoints = len(newZ)
        self.Z = newZ
        self.R = newR

    def uniformIncreasePointsNumber(self, n_points):
        Z2 = self.Z.pop()
        Z1 = self.Z.pop()
        R2 = self.R.pop()
        R1 = self.R.pop()
        for i in range(int(n_points + 1)):
            self.R.append((R2 - R1) / int(n_points) * i + R1)
            self.Z.append((Z2 - Z1) / int(n_points) * i + Z1)
        self.numberOfPoints = int(n_points)



from pynumeca.fine import fine
from pynumeca.igg import igg
from pynumeca.postprocessing import d3d, mf, plan, res
from pynumeca.preprocessing import bezier3d, bezier_geomturbo, turbo
from pynumeca.reader import geomturbo, numecaParser
from pynumeca.utils import boundaries, geometric, units

__all__ = [
    "fine",
    "igg",
    "d3d",
    "mf",
    "plan",
    "res",
    "bezier_geomturbo",
    "bezier3d",
    "turbo",
    "geomturbo",
    "numecaParser",
    "boundaries",
    "geometric",
    "units",
]



import os
import shutil
import subprocess

from pynumeca.constants import constants


def fine_run_from_mesh(
    source_iec_path: str,
    igg_path: str,
    output_iec_path: str,
    fine_version: str = constants.version,
    index: int = 1,
):
    shutil.copy(source_iec_path, output_iec_path)
    cmd = (
        f"fine{fine_version}"
        + " -print -batch"
        + " -project "
        + '"'
        + output_iec_path
        + '"'
        + " -mesh "
        + '"'
        + igg_path
        + '"'
        + " -index "
        + str(index)
    )

    print("Running command '" + cmd + "'")
    os.system(cmd)


def setup_parallel_computation(
    run_file_path: str,
    fine_version: str = constants.version,
    cores: int = 1,
    balancing_fine: float = constants.balancing_fine,
    memory_real: int = constants.memory_real,
    memory_int: int = constants.memory_int,
):
    cmd = (
        f"fine{fine_version}"
        + " -print -batch -partition"
        + " -computation "
        + '"'
        + run_file_path
        + '"'
        + " -nproc "
        + str(cores)
        + " -load_balancing "
        + str(balancing_fine)
        + " -nbint "
        + str(memory_int)
        + " -nbreal "
        + str(memory_real)
    )

    print("Running command '" + cmd + "'")
    os.system(cmd)


def run_parallel_computation(
    run_file_path: str, batch: bool = False, use_intel_mpi: bool = False
):
    run_file_path = os.path.split(run_file_path)[0]
    run_file_base_name = os.path.splitext(os.path.split(run_file_path)[1])[0]
    batch_file = os.path.join(run_file_path, run_file_base_name + ".batch")

    if use_intel_mpi:
        with open(batch_file, "a") as f:
            f.write("-intel -intelmpi ")

    if batch:
        cmd = batch_file
        print("Running command '" + cmd + "'")
        process = subprocess.Popen(cmd)
        return process.pid
    else:
        cmd = '"' + batch_file + '"'
        print("Running command '" + cmd + "'")
        os.system(cmd)
        return



import os
import re


def read_mf(mf_file: str) -> dict:
    """
    :param mf_file: .mf file path
    :return: dictionary representing the specified .plan file

    Reading a Numeca .mf, extracting main results and exporting them as a python dictionary

    """
    if not os.path.isfile(mf_file):
        raise FileNotFoundError(f"{mf_file} does not exist")

    # Open the file and read its contents
    with open(mf_file, "r") as f:
        contents = f.read()

    # Regular expressions to match the desired values
    inlet_pressure_pattern = "Static_pressure\s+([\d\.]+)"
    inlet_temperature_pattern = "Static_temperature\s+([\d\.]+)"
    mass_flow_pattern = "Mass_flow\s+([\d\.]+)"
    rotational_speed_pattern = "Rotational_speed\s+([\d\.\-]+)"
    inlet_absolute_total_pressure_pattern = "Absolute_total_pressure\s+([\d\.]+)"
    inlet_absolute_total_temperature_pattern = "Absolute_total_temperature\s+([\d\.]+)"
    absolute_total_pressure_ratio_pattern = "Absolute_total_pressure_ratio\s+([\d\.]+)"
    isentropic_efficiency_pattern = "Isentropic_efficiency\s+([\d\.]+)"

    # Find the values using regular expressions
    inlet_pressure = float(re.search(inlet_pressure_pattern, contents).group(1))
    inlet_temperature = float(re.search(inlet_temperature_pattern, contents).group(1))
    mass_flow = float(re.search(mass_flow_pattern, contents).group(1))
    rotational_speed = abs(
        float(re.search(rotational_speed_pattern, contents).group(1))
    )
    inlet_absolute_total_pressure = float(
        re.search(inlet_absolute_total_pressure_pattern, contents).group(1)
    )
    inlet_absolute_total_temperature = float(
        re.search(inlet_absolute_total_temperature_pattern, contents).group(1)
    )
    absolute_total_pressure_ratio = float(
        re.search(absolute_total_pressure_ratio_pattern, contents).group(1)
    )
    isentropic_efficiency = float(
        re.search(isentropic_efficiency_pattern, contents).group(1)
    )

    # Convert rotational speed from rad/s to RPM
    rotational_speed_rpm = rotational_speed * 9.549296596425384

    # Creating output dictionary
    out = {}
    out["Ps_in"] = inlet_pressure
    out["Ts_in"] = inlet_temperature
    out["Pt_in"] = inlet_absolute_total_pressure
    out["Tt_in"] = inlet_absolute_total_temperature
    out["Mass_flow"] = mass_flow
    out["Beta_tt"] = absolute_total_pressure_ratio
    out["Eta_tt"] = isentropic_efficiency
    out["Omega"] = rotational_speed_rpm

    return out



import numpy as np
from scipy.special import comb


def get_bezier_parameters(X, Y, Z, degree=3):
    """
    Least square qbezier fit using penrose pseudoinverse.

    Parameters:

    X: array of x data.
    Y: array of y data.
    Z: array of z data.

    degree: degree of the Bézier curve. 2 for quadratic, 3 for cubic.

    Based on https://stackoverflow.com/questions/12643079/b%C3%A9zier-curve-fitting-with-scipy
    and probably on the 1998 thesis by Tim Andrew Pastva, "Bézier Curve Fitting".
    """
    if degree < 1:
        raise ValueError("degree must be 1 or greater.")

    if len(X) != len(Y) != len(Z):
        raise ValueError("X and Y  and Z must be of the same length.")

    if len(X) < degree + 1:
        raise ValueError(
            f"There must be at least {degree + 1} points to "
            f"determine the parameters of a degree {degree} curve. "
            f"Got only {len(X)} points."
        )

    def bpoly(n, t, k):
        """Bernstein polynomial when a = 0 and b = 1."""
        return t**k * (1 - t) ** (n - k) * comb(n, k)
        # return comb(n, i) * ( t**(n-i) ) * (1 - t)**i

    def bmatrix(T):
        """Bernstein matrix for Bézier curves."""
        return np.matrix([[bpoly(degree, t, k) for k in range(degree + 1)] for t in T])

    def least_square_fit(points, M):
        M_ = np.linalg.pinv(M)
        return M_ * points

    T = np.linspace(0, 1, len(X))
    M = bmatrix(T)
    points = np.array(list(zip(X, Y, Z)))

    final = least_square_fit(points, M).tolist()
    final[0] = [X[0], Y[0], Z[0]]
    final[len(final) - 1] = [X[len(X) - 1], Y[len(Y) - 1], Z[len(Z) - 1]]
    return final


def bernstein_poly(i, n, t):
    """
    The Bernstein polynomial of n, i as a function of t
    """
    return comb(n, i) * (t ** (n - i)) * (1 - t) ** i


def bezier_curve(points, nTimes=50):
    """
    Given a set of control points, return the
    bezier curve defined by the control points.

    points should be a list of lists, or list of tuples
    such as [ [1,1,2],
              [2,3,1],
              [4,5,4], ..[Xn, Yn, Zn] ]

    nTimes is the number of time steps, defaults to 1000

    See http://processingjs.nihongoresources.com/bezierinfo/
    """

    nPoints = len(points)
    xPoints = np.array([p[0] for p in points])
    yPoints = np.array([p[1] for p in points])
    zPoints = np.array([p[2] for p in points])

    t = np.linspace(0.0, 1.0, nTimes)

    polynomial_array = np.array(
        [bernstein_poly(i, nPoints - 1, t) for i in range(0, nPoints)]
    )

    # TODO: why a is necessary to flip there?
    xvals = np.flip(np.dot(xPoints, polynomial_array))
    yvals = np.flip(np.dot(yPoints, polynomial_array))
    zvals = np.flip(np.dot(zPoints, polynomial_array))

    return xvals, yvals, zvals



import os

import joblib
import numpy as np
from pyfluids.fluids.fluid import Fluid, FluidsList

from pynumeca.postprocessing.mf import read_mf
from pynumeca.reader.numecaParser import numecaParser
from pynumeca.utils.boundaries import Boundaries
from pynumeca.utils.geometric import car2cil
from pynumeca.utils.units import convert_rotational_speed


class CentrifugalCompressor(object):
    def __init__(self, fluid: FluidsList, verbosity: bool = True):
        self.impeller = None
        self.splitter = None
        self.diffuser = None
        self.channel = None

        self.fluid = fluid

        self.rotating_speed = None
        self.external_diameter = None

        self.boundaries = None

        self.total_to_total_isentropic_efficiency = None
        self.total_to_total_pressure_ratio = None

        self.verbosity = verbosity

    def update_boundaries(self, m: float, pt_in: float, tt_in: float):
        self.boundaries = Boundaries(fluid=self.fluid, m=m, pt_in=pt_in, tt_in=tt_in)

    def load_boundaries_and_performances_from_mf(self, mf_path):
        mf_dict = read_mf(mf_path)
        self.rotating_speed = convert_rotational_speed(mf_dict["Omega"], to_rpm=False)
        self.total_to_total_isentropic_efficiency = mf_dict["Eta_tt"]
        self.total_to_total_pressure_ratio = mf_dict["Beta_tt"]

        self.update_boundaries(
            m=mf_dict["Mass_flow"], pt_in=mf_dict["Pt_in"], tt_in=mf_dict["Tt_in"]
        )

    def load_geometry_from_geomturbo(self, path: str):
        if not os.path.isfile(path):
            raise FileNotFoundError(f"{path} not found")

        parser = numecaParser()
        parser.load(filename=path)

        try:
            self.impeller = Blade(
                geometry=parser.exportNpyArray(0, 0)[0],
                periodicity=parser.get_row_periodicity(rowNumber=0),
            )
            self.external_diameter = find_external_diameter(self.impeller.geometry)

        except Exception as e:
            self.print("Cannot load impeller blade: ", e)

        try:
            self.splitter = Blade(
                geometry=parser.exportNpyArray(0, 1)[0],
                periodicity=parser.get_row_periodicity(rowNumber=0),
            )
        except Exception as e:
            self.print("Cannot load splitter blade: ", e)

        try:
            self.diffuser = Blade(
                geometry=parser.exportNpyArray(1, 0)[0],
                periodicity=parser.get_row_periodicity(rowNumber=1),
            )
        except Exception as e:
            self.print("Cannot load diffuser blade: ", e)

        try:
            hub, shroud = parser.exportZRNpyArraysList()
            self.channel = Channel(hub=hub, shroud=shroud)
        except Exception as e:
            self.print("Cannot load channel: ", e)

    def print(self, *args, **kwargs):
        print(*args, **kwargs) if self.verbosity else None

    def save(self, path: str):
        _, file_extension = os.path.splitext(path)
        if file_extension != ".bin":
            path += ".bin"

        with open(path, "wb") as f:
            joblib.dump(self, f)

        print(f"{self.__class__.__name__} saved to {path}")

    @staticmethod
    def load(path: str):
        if not os.path.exists(path):
            raise ValueError(f"{path} does not exist")
        return joblib.load(path)


class Blade(object):
    def __init__(
        self, geometry: np.ndarray, periodicity: int, is_te_blunt: bool = True
    ):
        self.geometry = geometry  # (2, S, N, 4)
        self.is_te_blunt = is_te_blunt
        self.periodicity = periodicity


class Channel(object):
    def __init__(self, hub: np.ndarray, shroud: np.ndarray):
        self.hub = hub
        self.shroud = shroud


def find_external_diameter(x):
    y = car2cil(x)
    return y[0, 0, -1, 0] * 2



import os.path
import signal
from typing import Optional

import numpy as np
import pandas as pd

from pynumeca.constants import constants
from pynumeca.fine import fine
from pynumeca.igg import igg
from pynumeca.postprocessing import mf


class Simulation(object):
    __res_cols = [
        "Iteration number",
        "Work unit",
        "CPU time",
        "Lift",
        "Drag",
        "Torque",
        "Qmax",
        "Tmax",
        "Mass flow in",
        "Mass flow out",
    ]

    def __init__(
        self,
        name: str = "Simulation",
        working_path: Optional[str] = None,
        geomturbo_path: Optional[str] = None,
        iec_path: Optional[str] = None,
        trb_path: Optional[str] = None,
        version: Optional[str] = None,
    ):
        if working_path is not None:
            self.working_path = working_path
        if geomturbo_path is not None:
            self.geomturbo_path = geomturbo_path
        if iec_path is not None:
            self.iec_path = iec_path
        if trb_path is not None:
            self.trb_path = trb_path
        if version is None:
            self.version = constants.version
        else:
            self.version = version

        self.name = name
        self.runfile = None
        self.runpid = None
        self.resfile = None

    def make_mesh(self):
        igg.a5_mesh_from_geomturbo(
            self.trb_path,
            self.geomturbo_path,
            # new_trb,
            os.path.join(self.working_path, self.name, self.name + ".igg"),
            igg_version=self.version,
        )

    def generate_run(self, index: int = 0):
        fine.fine_run_from_mesh(
            self.iec_path,
            os.path.join(self.working_path, self.name, self.name + ".igg"),
            os.path.join(self.working_path, self.name, self.name + ".iec"),
            index=index,
            fine_version=self.version,
        )

    @staticmethod
    def setup_parallel_computation(
        run_file_path: str, cores: int = 20, version: str = constants.version
    ):
        fine.setup_parallel_computation(
            run_file_path, cores=cores, fine_version=version
        )

    def run_pipeline(
        self, cores: int = 20, index: int = 0, use_intel_mpi: bool = False
    ):
        self.make_mesh()
        self.generate_run(index)
        all_subdirs = [
            os.path.join(self.working_path, self.name, d)
            for d in os.listdir(os.path.join(self.working_path, self.name))
            if os.path.isdir(os.path.join(self.working_path, self.name, d))
        ]
        all_runfiles = []
        for d in all_subdirs:
            for f in os.listdir(d):
                if os.path.isfile(os.path.join(d, f)) and ".run" in f:
                    all_runfiles.append(os.path.join(d, f))

        # print(all_runfiles)
        self.runfile = str(max(all_runfiles, key=os.path.getmtime))
        self.resfile = os.path.splitext(self.runfile)[0] + ".res"
        print("Latest run file:", self.runfile)
        self.setup_parallel_computation(self.runfile, cores, version=self.version)

        run_file_dir = os.path.split(self.runfile)[0]
        run_file_base_name = os.path.splitext(os.path.split(self.runfile)[1])[0]
        batch_file = os.path.join(run_file_dir, run_file_base_name + ".batch")

        self.run_parallel_computation(batch_file, use_intel_mpi)

    def read_resfile(self):
        with open(self.resfile, "r") as f:
            res_content = f.readlines()
        extract = []
        index_ = 0
        for line in res_content:
            if f" {index_} " in line:
                extract.append(
                    np.array(
                        [
                            item
                            for item in line.replace("\n", "").split(" ")
                            if item != ""
                        ],
                        dtype=np.float32,
                    )
                )
                index_ += 1
        dataframe = pd.DataFrame(extract, columns=self.__res_cols)
        return dataframe

    def read_mf(self):
        if self.runfile is not None:
            mf_path = os.path.splitext(self.runfile)[0] + ".mf"
            return mf.read_mf(mf_file=mf_path)

    def run_parallel_computation(self, batch_path: str, use_intel_mpi: bool = False):
        self.runpid = fine.run_parallel_computation(
            batch_path, use_intel_mpi=use_intel_mpi
        )

    def kill_parallel_computation(self):
        if self.runpid is not None:
            os.kill(self.runpid, signal.SIGTERM)

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        if self.working_path is not None:
            existing_titles = os.listdir(self.working_path)
            if value not in existing_titles:
                self._name = value
            else:
                index = 0
                while True:
                    index += 1
                    new_title = value + f"_{index}"
                    if new_title not in existing_titles:
                        break
                self._name = new_title

            print("Working directory: ", os.path.join(self.working_path, self._name))
            os.mkdir(os.path.join(self.working_path, self._name))

    @property
    def working_path(self):
        return self._working_path

    @working_path.setter
    def working_path(self, value):
        if os.path.exists(value):
            self._working_path = value
        else:
            try:
                os.mkdir(value)
                self._working_path = value
            except Exception as e:
                print(f'"{value}" error: {e}')

    @property
    def geomturbo_path(self):
        return self._geomturbo_path

    @geomturbo_path.setter
    def geomturbo_path(self, value):
        if os.path.exists(value):
            self._geomturbo_path = value
        else:
            print(f'"{value}" does not exists')

    @property
    def iec_path(self):
        return self._iec_path

    @iec_path.setter
    def iec_path(self, value):
        if os.path.exists(value):
            self._iec_path = value
        else:
            print(f'"{value}" does not exists')

    @property
    def trb_path(self):
        return self._trb_path

    @trb_path.setter
    def trb_path(self, value):
        if os.path.exists(value):
            self._trb_path = value
        else:
            print(f'"{value}" does not exists')



import re

from pynumeca.reader.numecaEntry import numecaEntry


class iecEntry(numecaEntry):
    def __init__(self, *args):
        super(iecEntry, self).__init__(*args)

    def parseSimpleEntry(self, string):
        if string == "\n":
            self.leadingSpaceString = ""
            self.key = ""
            self.keyTagSpaceString = ""
            self.tag = ""
            self.tagValueSpaceString = ""
            self.value = ""
            self.trailingSpaceString = "\n"
            return
        # splitPattern = '[\s]{2,}|(?<=[0-9])[\s]+(?=[0-9])|\b[\s]+$'
        splitPattern = "[\s]{2,}|(?<=\d)[\s]+(?=-?\d)|\b[\s]+$"
        tokenize = string.split()

        if len(tokenize) > 2 and "NI_BEGIN" not in string and "VERTEX" not in string:
            tokenize = re.split(splitPattern, string.strip())
            spaces = re.findall(splitPattern, string.strip())
        else:
            spaces = re.findall("\s+", string.strip())

        self.key = tokenize[0]
        self.leadingSpaceString = re.match(r"^(\s*)", string).group(1)

        self.trailingSpaceString = re.search(r"(\s*)$", string).group(1)
        # Verifica per i parametri, da estendere in futuro se necessario
        parameterCheck = re.findall("^[a-zA-z]*(,[A-Z])(,[0-9])", string.strip(" "))
        if len(tokenize) == 1:
            self.tag = ""
            self.value = ""
            self.keyTagSpaceString = ""
            self.tagValueSpaceString = ""
        elif string.lstrip(" ").startswith(("*", "#")):
            self.tag = ""
            self.value = ""
            self.keyTagSpaceString = ""
            self.tagValueSpaceString = ""
            self.key = string.lstrip(" ")[:-1]
        elif len(tokenize) == 2:
            self.tag = ""
            self.keyTagSpaceString = ""
            self.value = tokenize[1]
            self.tagValueSpaceString = spaces[0]
        elif len(tokenize) > 2:
            self.value = []
            self.tagValueSpaceString = []
            if (
                tokenize[1]
                .replace(".", "", 1)
                .replace("-", "", 2)
                .replace("E", "", 1)
                .replace("e", "", 1)
                .isnumeric()
            ):
                self.tag = ""
                self.keyTagSpaceString = ""
                for idx, token in enumerate(tokenize[1:]):
                    self.value.append(token)
                    self.tagValueSpaceString.append(spaces[idx])
            else:
                self.tag = tokenize[1]
                self.keyTagSpaceString = spaces[0]
                for idx, token in enumerate(tokenize[2:]):
                    self.value.append(token)
                    self.tagValueSpaceString.append(spaces[idx - 1])

        # if ("0.176764989579553" in string or "0.0032820608466864" in string):
        # print (tokenize)

        else:
            raise RuntimeError

    def outputString(self):
        valueString = ""
        if isinstance(self.value, str):
            valueString += self.tagValueSpaceString + self.value
        elif isinstance(self.value, list):
            for idx, token in enumerate(self.value):
                valueString += self.tagValueSpaceString[idx] + token
        return (
            self.leadingSpaceString
            + self.key
            + self.keyTagSpaceString
            + self.tag
            + valueString
            + self.trailingSpaceString
        )

    def __repr__(self):
        valueString = ""
        if isinstance(self.value, str):
            valueString += str(len(self.tagValueSpaceString)) + " " + self.value + "\n"
        elif isinstance(self.value, list):
            for idx, token in enumerate(self.value):
                valueString += (
                    str(len(self.tagValueSpaceString[idx])) + " " + token + "\n"
                )

        return (
            str(len(self.leadingSpaceString))
            + " "
            + self.key
            + "\n"
            + str(len(self.keyTagSpaceString))
            + " "
            + self.tag
            + "\n"
            + valueString
        )

    def __str__(self):
        valueString = ""
        if isinstance(self.value, str):
            valueString += self.value + "\n"
        elif isinstance(self.value, list):
            for idx, token in enumerate(self.value):
                valueString += self.tagValueSpaceString[idx] + token
            valueString += "\n"

        return self.key + "\n" + self.tag + "\n" + valueString



import numpy as np
import pandas as pd

__res_cols = [
    "Iteration number",
    "Work unit",
    "CPU time",
    "Lift",
    "Drag",
    "Torque",
    "Qmax",
    "Tmax",
    "Mass flow in",
    "Mass flow out",
]


def read_resfile(path: str):
    with open(path, "r") as f:
        res_content = f.readlines()
    extract = []
    index_ = 0
    for line in res_content:
        if f" {index_} " in line:
            extract.append(
                np.array(
                    [item for item in line.replace("\n", "").split(" ") if item != ""],
                    dtype=np.float32,
                )
            )
            index_ += 1
    dataframe = pd.DataFrame(extract, columns=__res_cols)
    return dataframe



from __future__ import annotations

import os
from typing import List, Optional

import joblib
import numpy as np
import plotly.graph_objects as go

from pynumeca.preprocessing.bezier3d import bezier_curve, get_bezier_parameters
from pynumeca.reader.numecaParser import numecaParser
from pynumeca.utils.geometric import fitting_metrics


class BezierChannel(object):
    def __init__(
        self,
        hub: Optional[list] = None,
        shroud: Optional[list] = None,
        control_points_hub: Optional[np.ndarray] = None,
        control_points_shroud: Optional[np.ndarray] = None,
        degree: int = 6,
        evaluation_points: int = 30,
        record_metrics: bool = False,
    ):
        self.hub = hub
        self.shroud = shroud
        self.control_points_hub = control_points_hub
        self.control_points_shroud = control_points_shroud
        self.degree = degree
        self.evaluation_points = evaluation_points

        self.fitted_hub = None
        self.fitted_shroud = None

        self.hub_degrees = None
        self.shroud_degrees = None

        self.fitting_metrics = []
        self.record_metrics = record_metrics

    def average_fitting_metrics(self):
        max_ = []
        mean_ = []

        for metric in self.fitting_metrics:
            max_.append(metric["max"])
            mean_.append(metric["mean"])

        max_ = np.array(max_)
        mean_ = np.array(mean_)

        return max_, mean_

    def __fit_with_bezier(self, segments_list: list):
        controls = []
        degrees = []

        for i, segment in enumerate(segments_list):
            if i != len(segments_list) - 1:
                #  TODO: why is necessry a tollerance there?
                # assert (np.round(segment[0, -1, :], 5) == np.round(segments_list[i + 1][0, 0, :], 5)).all(), f'Error in {self.__class__.__name__} bezier fitting'
                if not (
                    np.round(segment[0, -1, :], 8)
                    == np.round(segments_list[i + 1][0, 0, :], 8)
                ).all():
                    print("WARNING: channel __fit_with_bezier tollerance not respected")

            deg = 1 if len(segment[0]) == 2 else self.degree
            points = get_bezier_parameters(
                segment[0, :, 0], segment[0, :, 1], segment[0, :, 2], degree=deg
            )
            if self.record_metrics and segment.shape[1] > 2:
                xvals, yvals, zvals = bezier_curve(points, nTimes=100)
                self.fitting_metrics.append(
                    fitting_metrics(
                        segment[0, :, :3], np.vstack((xvals, yvals, zvals)).T
                    )
                )

            if i != len(segments_list) - 1:
                del points[-1]

            controls.append(points)
            degrees.append(deg)

        return np.vstack(controls), degrees

    def fit_channel_with_bezier(self):
        if self.hub is not None:
            self.control_points_hub, self.hub_degrees = self.__fit_with_bezier(self.hub)
        else:
            print(f"{self.__class__.__name__}.hub is None")

        if self.shroud is not None:
            self.control_points_shroud, self.shroud_degrees = self.__fit_with_bezier(
                self.shroud
            )
        else:
            print(f"{self.__class__.__name__}.shroud is None")

    def __get_from_control_points(self, control_points: np.ndarray, degrees: list):
        controls = []
        lower_bound: int = 0
        curve = []

        n_segments = len(degrees)

        for i in range(n_segments):
            if i != (n_segments - 1):
                upper_bound: int = (
                    lower_bound + 1 if degrees[i] == 1 else lower_bound + degrees[i]
                )

                controls.append(control_points[lower_bound:upper_bound])
            else:
                controls.append(control_points[lower_bound:])

            lower_bound = upper_bound

        for i in range(n_segments):
            if i != (n_segments - 1):
                # controls[i] = np.insert(controls[i], -1, controls[i+1][0], 0)
                controls[i] = np.concatenate([controls[i], controls[i + 1][:1]], axis=0)
                pass
            points = np.array(bezier_curve(controls[i], self.evaluation_points)).T
            curve.append(points)

        curve = np.concatenate(curve, axis=0)

        return curve

    def get_channel_from_control_points(self):
        self.fitted_hub = self.__get_from_control_points(
            self.control_points_hub, self.hub_degrees
        )
        self.fitted_shroud = self.__get_from_control_points(
            self.control_points_shroud, self.shroud_degrees
        )

    def draw_channel(self, show: bool = False, width: int = 1000, height: int = 800):
        fig = go.Figure(
            data=[
                go.Scatter3d(
                    x=self.control_points_hub[:, 0],
                    y=self.control_points_hub[:, 1],
                    z=self.control_points_hub[:, 2],
                    mode="markers",
                    marker=dict(size=5),
                    name="Control Points Hub",
                ),
                go.Scatter3d(
                    x=self.fitted_hub[:, 0],
                    y=self.fitted_hub[:, 1],
                    z=self.fitted_hub[:, 2],
                    mode="markers",
                    marker=dict(size=5),
                    name="Fitted curve Hub",
                ),
                go.Scatter3d(
                    x=self.control_points_shroud[:, 0],
                    y=self.control_points_shroud[:, 1],
                    z=self.control_points_shroud[:, 2],
                    mode="markers",
                    marker=dict(size=5),
                    name="Control Points Shroud",
                ),
                go.Scatter3d(
                    x=self.fitted_shroud[:, 0],
                    y=self.fitted_shroud[:, 1],
                    z=self.fitted_shroud[:, 2],
                    mode="markers",
                    marker=dict(size=5),
                    name="Fitted curve Shroud",
                ),
            ],
            layout=go.Layout(
                width=width,
                height=height,
            ),
        )
        fig.show() if show else None
        return fig


class BezierBlade(object):
    __LE_POINTS_NUMBER = 15
    __TE_POINTS_NUMBER = 15

    def __init__(
        self,
        numpy_blade: Optional[np.ndarray] = None,
        control_points: Optional[np.ndarray] = None,
        bezier_degree: int | list = 6,
        evaluation_points: int | list = 100,
        is_blunt_te: bool = True,
        split_edges: bool = False,
        record_metrics: bool = False,
    ):
        self.numpy_blade = numpy_blade  # (S, s, n, f)
        self.control_points = control_points
        self.is_blunt_te = is_blunt_te
        self.split_edges = split_edges

        self.fitted_numpy_blade = None

        self.set_bezier_degree(bezier_degree)
        self.set_evaluation_points(evaluation_points)

        self.fitting_metrics = []
        self.record_metrics = record_metrics

    def average_fitting_metrics(self):
        max_ = []
        mean_ = []

        for metric in self.fitting_metrics:
            max_.append(metric["max"])
            mean_.append(metric["mean"])

        max_ = np.array(max_)
        mean_ = np.array(mean_)

        return max_, mean_

    def set_bezier_degree(self, bezier_degree: int | list):
        if isinstance(bezier_degree, int):
            if self.split_edges:
                if self.is_blunt_te:
                    self.__bezier_degree = [bezier_degree for _ in range(4)]
                else:
                    self.__bezier_degree = [bezier_degree for _ in range(6)]
            else:
                if self.is_blunt_te:
                    self.__bezier_degree = [bezier_degree for _ in range(3)]
                else:
                    self.__bezier_degree = [bezier_degree for _ in range(4)]
        else:
            self.__bezier_degree = bezier_degree

    def get_bezier_degree(self) -> list:
        return self.__bezier_degree

    def set_evaluation_points(self, evaluation_points: int | list):
        if isinstance(evaluation_points, int):
            if self.split_edges:
                if self.is_blunt_te:
                    self.__evaluation_points = [evaluation_points for _ in range(4)]
                else:
                    self.__evaluation_points = [evaluation_points for _ in range(6)]
            else:
                if self.is_blunt_te:
                    self.__evaluation_points = [evaluation_points for _ in range(3)]
                else:
                    self.__evaluation_points = [evaluation_points for _ in range(4)]
        else:
            self.__evaluation_points = evaluation_points

    def get_evaluation_points(self) -> list:
        return self.__evaluation_points

    def set_le_points_number(self, number: int) -> None:
        if isinstance(number, int):
            self.__LE_POINTS_NUMBER = number
        else:
            raise TypeError(f"number must be an integer. number type is {type(number)}")

    def set_te_points_number(self, number: int) -> None:
        if isinstance(number, int):
            self.__TE_POINTS_NUMBER = number
        else:
            raise TypeError(f"number must be an integer. number type is {type(number)}")

    def __fit_section_with_bezier(
        self, section: Optional[np.ndarray] = None
    ) -> np.ndarray:
        # section (S, n, f)

        assert (
            section[0, 0, :] == section[1, -1, :]
        ).all(), "The last point of the suction side point and the pressure side first one must be identical"

        if self.split_edges:
            if self.is_blunt_te:
                ss = section[1, : -self.__LE_POINTS_NUMBER, :]
                ss_le = section[1, -self.__LE_POINTS_NUMBER - 1 :, :]

                ps_le = section[0, : self.__LE_POINTS_NUMBER + 1, :]
                ps = section[0, self.__LE_POINTS_NUMBER :, :]

                ss_control_points = np.array(
                    get_bezier_parameters(
                        ss[:, 0], ss[:, 1], ss[:, 2], degree=self.__bezier_degree[0]
                    )
                )

                if self.record_metrics:
                    xvals, yvals, zvals = bezier_curve(ss_control_points, nTimes=100)
                    self.fitting_metrics.append(
                        fitting_metrics(ss[..., :3], np.vstack((xvals, yvals, zvals)).T)
                    )

                ss_le_control_points = np.array(
                    get_bezier_parameters(
                        ss_le[:, 0],
                        ss_le[:, 1],
                        ss_le[:, 2],
                        degree=self.__bezier_degree[1],
                    )
                )

                if self.record_metrics:
                    xvals, yvals, zvals = bezier_curve(ss_le_control_points, nTimes=100)
                    self.fitting_metrics.append(
                        fitting_metrics(
                            ss_le[..., :3], np.vstack((xvals, yvals, zvals)).T
                        )
                    )

                ps_le_control_points = np.array(
                    get_bezier_parameters(
                        ps_le[:, 0],
                        ps_le[:, 1],
                        ps_le[:, 2],
                        degree=self.__bezier_degree[2],
                    )
                )

                if self.record_metrics:
                    xvals, yvals, zvals = bezier_curve(ps_le_control_points, nTimes=100)
                    self.fitting_metrics.append(
                        fitting_metrics(
                            ps_le[..., :3], np.vstack((xvals, yvals, zvals)).T
                        )
                    )

                ps_control_points = np.array(
                    get_bezier_parameters(
                        ps[:, 0], ps[:, 1], ps[:, 2], degree=self.__bezier_degree[3]
                    )
                )

                if self.record_metrics:
                    xvals, yvals, zvals = bezier_curve(ps_control_points, nTimes=100)
                    self.fitting_metrics.append(
                        fitting_metrics(ps[..., :3], np.vstack((xvals, yvals, zvals)).T)
                    )

                assert (
                    (ss_control_points[-1] == ss_le_control_points[0]).all()
                    and (ss_le_control_points[-1] == ps_le_control_points[0]).all()
                    and (ps_le_control_points[-1] == ps_control_points[0]).all()
                ), "Error: please check fit_section_with_bezier function"

                #  As a rule of thumb, we remove the first control point of the second curve in a joint for ML purposes
                ss_le_control_points = np.delete(ss_le_control_points, 0, axis=0)
                ps_le_control_points = np.delete(ps_le_control_points, 0, axis=0)
                ps_control_points = np.delete(ps_control_points, 0, axis=0)

                return np.concatenate(
                    [
                        ss_control_points,
                        ss_le_control_points,
                        ps_le_control_points,
                        ps_control_points,
                    ],
                    axis=0,
                )

            else:
                ss_te = section[1, : self.__TE_POINTS_NUMBER, :]
                ss = section[
                    1, self.__TE_POINTS_NUMBER - 1 : -self.__LE_POINTS_NUMBER, :
                ]
                ss_le = section[1, -self.__LE_POINTS_NUMBER - 1 :, :]
                ps_le = section[0, : self.__LE_POINTS_NUMBER + 1, :]
                ps = section[0, self.__LE_POINTS_NUMBER : -self.__TE_POINTS_NUMBER, :]
                ps_te = section[0, -self.__TE_POINTS_NUMBER - 1 :, :]

                ss_te_control_points = np.array(
                    get_bezier_parameters(
                        ss_te[:, 0],
                        ss_te[:, 1],
                        ss_te[:, 2],
                        degree=self.__bezier_degree[0],
                    )
                )

                if self.record_metrics:
                    xvals, yvals, zvals = bezier_curve(ss_te_control_points, nTimes=100)
                    self.fitting_metrics.append(
                        fitting_metrics(
                            ss_te[..., :3], np.vstack((xvals, yvals, zvals)).T
                        )
                    )

                ss_control_points = np.array(
                    get_bezier_parameters(
                        ss[:, 0], ss[:, 1], ss[:, 2], degree=self.__bezier_degree[1]
                    )
                )

                if self.record_metrics:
                    xvals, yvals, zvals = bezier_curve(ss_control_points, nTimes=100)
                    self.fitting_metrics.append(
                        fitting_metrics(ss[..., :3], np.vstack((xvals, yvals, zvals)).T)
                    )

                ss_le_control_points = np.array(
                    get_bezier_parameters(
                        ss_le[:, 0],
                        ss_le[:, 1],
                        ss_le[:, 2],
                        degree=self.__bezier_degree[2],
                    )
                )

                if self.record_metrics:
                    xvals, yvals, zvals = bezier_curve(ss_le_control_points, nTimes=100)
                    self.fitting_metrics.append(
                        fitting_metrics(
                            ss_le[..., :3], np.vstack((xvals, yvals, zvals)).T
                        )
                    )

                ps_le_control_points = np.array(
                    get_bezier_parameters(
                        ps_le[:, 0],
                        ps_le[:, 1],
                        ps_le[:, 2],
                        degree=self.__bezier_degree[3],
                    )
                )

                if self.record_metrics:
                    xvals, yvals, zvals = bezier_curve(ps_le_control_points, nTimes=100)
                    self.fitting_metrics.append(
                        fitting_metrics(
                            ps_le[..., :3], np.vstack((xvals, yvals, zvals)).T
                        )
                    )

                ps_control_points = np.array(
                    get_bezier_parameters(
                        ps[:, 0], ps[:, 1], ps[:, 2], degree=self.__bezier_degree[4]
                    )
                )

                if self.record_metrics:
                    xvals, yvals, zvals = bezier_curve(ps_control_points, nTimes=100)
                    self.fitting_metrics.append(
                        fitting_metrics(ps[..., :3], np.vstack((xvals, yvals, zvals)).T)
                    )

                ps_te_control_points = np.array(
                    get_bezier_parameters(
                        ps_te[:, 0],
                        ps_te[:, 1],
                        ps_te[:, 2],
                        degree=self.__bezier_degree[5],
                    )
                )

                if self.record_metrics:
                    xvals, yvals, zvals = bezier_curve(ps_te_control_points, nTimes=100)
                    self.fitting_metrics.append(
                        fitting_metrics(
                            ps_te[..., :3], np.vstack((xvals, yvals, zvals)).T
                        )
                    )

                assert (
                    (ss_te_control_points[-1] == ss_control_points[0]).all()
                    and (ss_control_points[-1] == ss_le_control_points[0]).all()
                    and (ss_le_control_points[-1] == ps_le_control_points[0]).all()
                    and (ps_le_control_points[-1] == ps_control_points[0]).all()
                    and (ps_control_points[-1] == ps_te_control_points[0]).all()
                    and (ps_te_control_points[-1] == ss_te_control_points[0]).all()
                ), "Error: please check fit_section_with_bezier function"

                ss_te_control_points = np.delete(ss_te_control_points, 0, axis=0)
                ss_control_points = np.delete(ss_control_points, 0, axis=0)
                ss_le_control_points = np.delete(ss_le_control_points, 0, axis=0)
                ps_le_control_points = np.delete(ps_le_control_points, 0, axis=0)
                ps_control_points = np.delete(ps_control_points, 0, axis=0)
                ps_te_control_points = np.delete(ps_te_control_points, 0, axis=0)

                return np.concatenate(
                    [
                        ss_te_control_points,
                        ss_control_points,
                        ss_le_control_points,
                        ps_le_control_points,
                        ps_control_points,
                        ps_te_control_points,
                    ],
                    axis=0,
                )

        else:
            if self.is_blunt_te:
                ss = section[1, : -self.__LE_POINTS_NUMBER, :]

                ss_le = section[1, -self.__LE_POINTS_NUMBER - 1 :, :]
                ps_le = section[0, 1 : self.__LE_POINTS_NUMBER + 1, :]

                ps = section[0, self.__LE_POINTS_NUMBER :, :]

                le = np.concatenate((ss_le, ps_le), axis=0)

                ss_control_points = np.array(
                    get_bezier_parameters(
                        ss[:, 0], ss[:, 1], ss[:, 2], degree=self.__bezier_degree[0]
                    )
                )

                if self.record_metrics:
                    xvals, yvals, zvals = bezier_curve(ss_control_points, nTimes=100)
                    self.fitting_metrics.append(
                        fitting_metrics(ss[..., :3], np.vstack((xvals, yvals, zvals)).T)
                    )

                le_control_points = np.array(
                    get_bezier_parameters(
                        le[:, 0], le[:, 1], le[:, 2], degree=self.__bezier_degree[1]
                    )
                )

                if self.record_metrics:
                    xvals, yvals, zvals = bezier_curve(le_control_points, nTimes=100)
                    self.fitting_metrics.append(
                        fitting_metrics(le[..., :3], np.vstack((xvals, yvals, zvals)).T)
                    )

                ps_control_points = np.array(
                    get_bezier_parameters(
                        ps[:, 0], ps[:, 1], ps[:, 2], degree=self.__bezier_degree[2]
                    )
                )

                if self.record_metrics:
                    xvals, yvals, zvals = bezier_curve(ps_control_points, nTimes=100)
                    self.fitting_metrics.append(
                        fitting_metrics(ps[..., :3], np.vstack((xvals, yvals, zvals)).T)
                    )

                assert (ss_control_points[-1] == le_control_points[0]).all() and (
                    le_control_points[-1] == ps_control_points[0]
                ).all(), "Error: please check fit_section_with_bezier function"

                #  As a rule of thumb, we remove the first control point of the second curve in a joint for ML purposes
                le_control_points = np.delete(le_control_points, 0, axis=0)
                ps_control_points = np.delete(ps_control_points, 0, axis=0)

                return np.concatenate(
                    [ss_control_points, le_control_points, ps_control_points], axis=0
                )

            else:
                ss_te = section[1, 1 : self.__TE_POINTS_NUMBER, :]

                ss = section[
                    1, self.__TE_POINTS_NUMBER - 1 : -self.__LE_POINTS_NUMBER, :
                ]

                ss_le = section[1, -self.__LE_POINTS_NUMBER - 1 :, :]
                ps_le = section[0, 1 : self.__LE_POINTS_NUMBER + 1, :]

                ps = section[0, self.__LE_POINTS_NUMBER : -self.__TE_POINTS_NUMBER, :]

                ps_te = section[0, -self.__TE_POINTS_NUMBER - 1 :, :]

                le = np.concatenate((ss_le, ps_le), axis=0)
                te = np.concatenate((ps_te, ss_te), axis=0)

                ss_control_points = np.array(
                    get_bezier_parameters(
                        ss[:, 0], ss[:, 1], ss[:, 2], degree=self.__bezier_degree[0]
                    )
                )

                if self.record_metrics:
                    xvals, yvals, zvals = bezier_curve(ss_control_points, nTimes=100)
                    self.fitting_metrics.append(
                        fitting_metrics(ss[..., :3], np.vstack((xvals, yvals, zvals)).T)
                    )

                le_control_points = np.array(
                    get_bezier_parameters(
                        le[:, 0], le[:, 1], le[:, 2], degree=self.__bezier_degree[1]
                    )
                )

                if self.record_metrics:
                    xvals, yvals, zvals = bezier_curve(le_control_points, nTimes=100)
                    self.fitting_metrics.append(
                        fitting_metrics(le[..., :3], np.vstack((xvals, yvals, zvals)).T)
                    )

                ps_control_points = np.array(
                    get_bezier_parameters(
                        ps[:, 0], ps[:, 1], ps[:, 2], degree=self.__bezier_degree[2]
                    )
                )

                if self.record_metrics:
                    xvals, yvals, zvals = bezier_curve(ps_control_points, nTimes=100)
                    self.fitting_metrics.append(
                        fitting_metrics(ps[..., :3], np.vstack((xvals, yvals, zvals)).T)
                    )

                te_control_points = np.array(
                    get_bezier_parameters(
                        te[:, 0], te[:, 1], te[:, 2], degree=self.__bezier_degree[3]
                    )
                )

                if self.record_metrics:
                    xvals, yvals, zvals = bezier_curve(te_control_points, nTimes=100)
                    self.fitting_metrics.append(
                        fitting_metrics(te[..., :3], np.vstack((xvals, yvals, zvals)).T)
                    )

                assert (
                    (ss_control_points[-1] == le_control_points[0]).all()
                    and (le_control_points[-1] == ps_control_points[0]).all()
                    and (ps_control_points[-1] == te_control_points[0]).all()
                    and (te_control_points[-1] == ss_control_points[0]).all()
                ), "Error: please check fit_section_with_bezier function"

                ss_control_points = np.delete(ss_control_points, 0, axis=0)
                le_control_points = np.delete(le_control_points, 0, axis=0)
                ps_control_points = np.delete(ps_control_points, 0, axis=0)
                te_control_points = np.delete(te_control_points, 0, axis=0)

                return np.concatenate(
                    [
                        ss_control_points,
                        le_control_points,
                        ps_control_points,
                        te_control_points,
                    ],
                    axis=0,
                )

    def fit_blade_with_bezier(self):
        if self.numpy_blade is not None:
            control_points = []
            for i in range(self.numpy_blade.shape[1]):
                control_points.append(
                    self.__fit_section_with_bezier(self.numpy_blade[:, i, :, :3])
                )

            control_points = np.array(control_points)
            self.control_points = control_points
            return control_points

        else:
            raise ValueError(f"{self.__class__.__name__}.numpy_blade is none")

    def __get_section_from_control_points(self, control_points: np.ndarray):
        if self.split_edges:
            if self.is_blunt_te:
                # (N, 3)
                ss_control_points = control_points[: self.__bezier_degree[0] + 1]

                ss_le_control_points = control_points[
                    self.__bezier_degree[0]
                    + 1 : self.__bezier_degree[0]
                    + 1
                    + self.__bezier_degree[1]
                ]
                ss_le_control_points = np.insert(
                    ss_le_control_points, 0, ss_control_points[-1], axis=0
                )

                ps_le_control_points = control_points[
                    self.__bezier_degree[0]
                    + 1
                    + self.__bezier_degree[1] : self.__bezier_degree[0]
                    + 1
                    + self.__bezier_degree[1]
                    + self.__bezier_degree[2]
                ]
                ps_le_control_points = np.insert(
                    ps_le_control_points, 0, ss_le_control_points[-1], axis=0
                )

                ps_control_points = control_points[
                    self.__bezier_degree[0]
                    + 1
                    + self.__bezier_degree[1]
                    + self.__bezier_degree[2] :
                ]
                ps_control_points = np.insert(
                    ps_control_points, 0, ps_le_control_points[-1], axis=0
                )

                ss = np.array(
                    bezier_curve(ss_control_points, nTimes=self.__evaluation_points[0])
                ).T
                ss_le = np.array(
                    bezier_curve(
                        ss_le_control_points, nTimes=self.__evaluation_points[1]
                    )
                ).T
                ps_le = np.array(
                    bezier_curve(
                        ps_le_control_points, nTimes=self.__evaluation_points[2]
                    )
                ).T
                ps = np.array(
                    bezier_curve(ps_control_points, nTimes=self.__evaluation_points[3])
                ).T

                return np.concatenate([ss, ss_le, ps_le, ps], axis=0)
            else:
                ss_te_control_points = control_points[: self.__bezier_degree[0]]

                ss_control_points = control_points[
                    self.__bezier_degree[0] : self.__bezier_degree[0]
                    + self.__bezier_degree[1]
                ]
                ss_control_points = np.insert(
                    ss_control_points, 0, ss_te_control_points[-1], axis=0
                )

                ss_le_control_points = control_points[
                    self.__bezier_degree[0]
                    + self.__bezier_degree[1] : self.__bezier_degree[0]
                    + self.__bezier_degree[1]
                    + self.__bezier_degree[2]
                ]
                ss_le_control_points = np.insert(
                    ss_le_control_points, 0, ss_control_points[-1], axis=0
                )

                ps_le_control_points = control_points[
                    self.__bezier_degree[0]
                    + self.__bezier_degree[1]
                    + self.__bezier_degree[2] : self.__bezier_degree[0]
                    + self.__bezier_degree[1]
                    + self.__bezier_degree[2]
                    + self.__bezier_degree[3]
                ]
                ps_le_control_points = np.insert(
                    ps_le_control_points, 0, ss_le_control_points[-1], axis=0
                )

                ps_control_points = control_points[
                    self.__bezier_degree[0]
                    + self.__bezier_degree[1]
                    + self.__bezier_degree[2]
                    + self.__bezier_degree[3] : self.__bezier_degree[0]
                    + self.__bezier_degree[1]
                    + self.__bezier_degree[2]
                    + self.__bezier_degree[3]
                    + self.__bezier_degree[4]
                ]
                ps_control_points = np.insert(
                    ps_control_points, 0, ps_le_control_points[-1], axis=0
                )

                ps_te_control_points = control_points[
                    self.__bezier_degree[0]
                    + self.__bezier_degree[1]
                    + self.__bezier_degree[2]
                    + self.__bezier_degree[3]
                    + self.__bezier_degree[4] :
                ]
                ps_te_control_points = np.insert(
                    ps_te_control_points, 0, ps_control_points[-1], axis=0
                )

                ss_te_control_points = np.insert(
                    ss_te_control_points, 0, ps_te_control_points[-1], axis=0
                )

                ss_te = np.array(
                    bezier_curve(
                        ss_te_control_points, nTimes=self.__evaluation_points[0]
                    )
                ).T
                ss = np.array(
                    bezier_curve(ss_control_points, nTimes=self.__evaluation_points[1])
                ).T
                ss_le = np.array(
                    bezier_curve(
                        ss_le_control_points, nTimes=self.__evaluation_points[2]
                    )
                ).T
                ps_le = np.array(
                    bezier_curve(
                        ps_le_control_points, nTimes=self.__evaluation_points[3]
                    )
                ).T
                ps = np.array(
                    bezier_curve(ps_control_points, nTimes=self.__evaluation_points[4])
                ).T
                ps_te = np.array(
                    bezier_curve(
                        ps_te_control_points, nTimes=self.__evaluation_points[5]
                    )
                ).T

                return np.concatenate([ss_te, ss, ss_le, ps_le, ps, ps_te], axis=0)
        else:
            if self.is_blunt_te:
                ss_control_points = control_points[: self.__bezier_degree[0] + 1]

                le_control_points = control_points[
                    self.__bezier_degree[0]
                    + 1 : self.__bezier_degree[0]
                    + 1
                    + self.__bezier_degree[1]
                ]
                le_control_points = np.insert(
                    le_control_points, 0, ss_control_points[-1], axis=0
                )

                ps_control_points = control_points[
                    self.__bezier_degree[0] + 1 + self.__bezier_degree[1] :
                ]
                ps_control_points = np.insert(
                    ps_control_points, 0, le_control_points[-1], axis=0
                )

                ss = np.array(
                    bezier_curve(ss_control_points, nTimes=self.__evaluation_points[0])
                ).T
                le = np.array(
                    bezier_curve(le_control_points, nTimes=self.__evaluation_points[1])
                ).T
                ps = np.array(
                    bezier_curve(ps_control_points, nTimes=self.__evaluation_points[2])
                ).T

                return np.concatenate([ss, le, ps], axis=0)
            else:
                ss_control_points = control_points[: self.__bezier_degree[0]]

                le_control_points = control_points[
                    self.__bezier_degree[0] : self.__bezier_degree[0]
                    + self.__bezier_degree[1]
                ]
                le_control_points = np.insert(
                    le_control_points, 0, ss_control_points[-1], axis=0
                )

                ps_control_points = control_points[
                    self.__bezier_degree[0]
                    + self.__bezier_degree[1] : self.__bezier_degree[0]
                    + self.__bezier_degree[1]
                    + self.__bezier_degree[2]
                ]
                ps_control_points = np.insert(
                    ps_control_points, 0, le_control_points[-1], axis=0
                )

                te_control_points = control_points[
                    self.__bezier_degree[0]
                    + self.__bezier_degree[1]
                    + self.__bezier_degree[2] :
                ]
                te_control_points = np.insert(
                    te_control_points, 0, ps_control_points[-1], axis=0
                )

                ss_control_points = np.insert(
                    ss_control_points, 0, te_control_points[-1], axis=0
                )

                ss = np.array(
                    bezier_curve(ss_control_points, nTimes=self.__evaluation_points[0])
                ).T
                le = np.array(
                    bezier_curve(le_control_points, nTimes=self.__evaluation_points[1])
                ).T
                ps = np.array(
                    bezier_curve(ps_control_points, nTimes=self.__evaluation_points[2])
                ).T
                te = np.array(
                    bezier_curve(te_control_points, nTimes=self.__evaluation_points[3])
                ).T

                return np.concatenate([ss, le, ps, te], axis=0)

    def get_blade_from_control_points(self):
        if self.control_points is not None:
            sections = []
            for i in range(self.control_points.shape[0]):
                sections.append(
                    self.__get_section_from_control_points(self.control_points[i])
                )

            self.fitted_numpy_blade = np.array(sections)
            return self.fitted_numpy_blade

        else:
            raise ValueError(f"{self.__class__.__name__}.control_points is none")

    def add_noise(self, sigma: float = 0.01):
        if self.control_points is not None:
            noise = (np.random.rand(*self.control_points.shape) - 0.5) * sigma
            self.control_points = self.control_points + noise
        else:
            print(f"{self.__class__.__name__}.control points are None")

    def draw_blade(self, show: bool = False, width: int = 1000, height: int = 800):
        data = []
        if self.numpy_blade is not None:
            bld = self.numpy_blade[:, :, :, :3].reshape(-1, 3)
            data.append(
                go.Scatter3d(
                    x=bld[:, 0],
                    y=bld[:, 1],
                    z=bld[:, 2],
                    mode="markers",
                    marker=dict(size=5),
                    name="Original Blade",
                ),
            )

        if self.control_points is not None:
            controls = self.control_points.reshape(-1, 3)
            data.append(
                go.Scatter3d(
                    x=controls[:, 0],
                    y=controls[:, 1],
                    z=controls[:, 2],
                    mode="markers",
                    marker=dict(size=5),
                    name="Control Points",
                ),
            )

        if self.fitted_numpy_blade is not None:
            bld = self.fitted_numpy_blade.reshape(-1, 3)
            data.append(
                go.Scatter3d(
                    x=bld[:, 0],
                    y=bld[:, 1],
                    z=bld[:, 2],
                    mode="markers",
                    marker=dict(size=5),
                    name="Fitted Blade",
                ),
            )

        fig = go.Figure(
            data=data,
            layout=go.Layout(
                width=width,
                height=height,
            ),
        )
        fig.show() if show else None
        return fig


class BezierCompressor(object):
    def __init__(
        self,
        geomturbo_path: str,
        bezier_degree: list | List[list, list, list],
        is_blunt_te: list,
        evaluation_points: list | List[list, list, list],
        split_edges: list,
        is_main_blade_active: bool = True,
        is_splitter_active: bool = True,
        is_diffuser_active: bool = True,
        is_channel_active: bool = True,
        channel_degree: int = 6,
        channel_evaluation_points: int = 30,
        record_metrics: bool = False,
    ):
        if not isinstance(bezier_degree[0], list):
            bezier_degree = [bezier_degree for _ in range(3)]
        if not isinstance(evaluation_points[0], list):
            evaluation_points = [evaluation_points for _ in range(3)]

        self.geomturbo_path = geomturbo_path
        self.bezier_degree = bezier_degree
        self.evaluation_points = evaluation_points
        self.is_blunt_te = is_blunt_te
        self.is_main_blade_active = is_main_blade_active
        self.is_splitter_active = is_splitter_active
        self.is_diffuser_active = is_diffuser_active
        self.is_channel_active = is_channel_active
        self.channel_evaluation_points = channel_evaluation_points

        self.main_blade = self.splitter = self.diffuser = None

        if is_main_blade_active:
            self.main_blade = BezierBlade(
                bezier_degree=self.bezier_degree[0],
                is_blunt_te=self.is_blunt_te[0],
                evaluation_points=self.evaluation_points[0],
                split_edges=split_edges[0],
                record_metrics=record_metrics,
            )
        if is_splitter_active:
            self.splitter = BezierBlade(
                bezier_degree=self.bezier_degree[1],
                is_blunt_te=self.is_blunt_te[1],
                evaluation_points=self.evaluation_points[1],
                split_edges=split_edges[1],
                record_metrics=record_metrics,
            )
        if is_diffuser_active:
            self.diffuser = BezierBlade(
                bezier_degree=self.bezier_degree[2],
                is_blunt_te=self.is_blunt_te[2],
                evaluation_points=self.evaluation_points[2],
                split_edges=split_edges[2],
                record_metrics=record_metrics,
            )
        if is_channel_active:
            self.channel = BezierChannel(
                degree=channel_degree,
                evaluation_points=channel_evaluation_points,
                record_metrics=record_metrics,
            )

        self.main_blade_control_points = None
        self.splitter_control_points = None
        self.diffuser_control_points = None

    def set_le_points_number(self, points_number: int | List[int, int, int]) -> None:
        if isinstance(points_number, int):
            points_number = [points_number for _ in range(3)]
        self.main_blade.set_le_points_number(
            points_number[0]
        ) if self.is_main_blade_active else None
        self.splitter.set_le_points_number(
            points_number[1]
        ) if self.is_splitter_active else None
        self.diffuser.set_le_points_number(
            points_number[2]
        ) if self.is_diffuser_active else None

    def set_te_points_number(self, points_number: int | List[int, int, int]) -> None:
        if isinstance(points_number, int):
            points_number = [points_number for _ in range(3)]
        self.main_blade.set_te_points_number(
            points_number[0]
        ) if self.is_main_blade_active else None
        self.splitter.set_te_points_number(
            points_number[1]
        ) if self.is_splitter_active else None
        self.diffuser.set_te_points_number(
            points_number[2]
        ) if self.is_diffuser_active else None

    def load_compressor_from_file(self):
        if os.path.exists(self.geomturbo_path):
            inputFile = numecaParser()
            inputFile.load(self.geomturbo_path)

            if self.is_main_blade_active:
                self.main_blade.numpy_blade = inputFile.exportNpyArray(0, 0)[0]
            if self.is_splitter_active:
                self.splitter.numpy_blade = inputFile.exportNpyArray(0, 1)[0]
            if self.is_diffuser_active:
                self.diffuser.numpy_blade = inputFile.exportNpyArray(1, 0)[0]
            if self.is_channel_active:
                (
                    self.channel.hub,
                    self.channel.shroud,
                ) = inputFile.exportZRNpyArraysList()

        else:
            raise FileNotFoundError(f"{self.geomturbo_path} does not exist")

    def set_control_points(self, tag: str, control_points: np.ndarray):
        if tag == "main_blade":
            if self.is_main_blade_active:
                self.main_blade.control_points = control_points
                self.main_blade_control_points = control_points
            else:
                raise AttributeError("Main blade is not active")
        elif tag == "splitter":
            if self.is_splitter_active:
                self.splitter.control_points = control_points
                self.splitter_control_points = control_points
            else:
                raise AttributeError("Splitter is not active")
        elif tag == "diffuser":
            if self.is_diffuser_active:
                self.diffuser.control_points = control_points
                self.diffuser_control_points = control_points
            else:
                raise AttributeError("Diffuser is not active")
        elif tag == "hub":
            if self.is_channel_active:
                self.channel.control_points_hub = control_points
            else:
                raise AttributeError("Channel is not active")
        elif tag == "shroud":
            if self.is_channel_active:
                self.channel.control_points_shroud = control_points
            else:
                raise AttributeError("Channel is not active")

        else:
            print(
                f'Invalid tag: {tag}. Valid tags: "main_blade", "splitter", "diffuser", "hub", "shroud"'
            )

    def fit_compressor_with_bezier(self):
        self.main_blade_control_points = (
            self.main_blade.fit_blade_with_bezier()
            if self.is_main_blade_active
            else None
        )
        self.splitter_control_points = (
            self.splitter.fit_blade_with_bezier() if self.is_splitter_active else None
        )
        self.diffuser_control_points = (
            self.diffuser.fit_blade_with_bezier() if self.is_diffuser_active else None
        )
        self.channel.fit_channel_with_bezier() if self.is_channel_active else None

    def get_compressor_from_control_points(self):
        self.main_blade.get_blade_from_control_points() if self.is_main_blade_active else None
        self.splitter.get_blade_from_control_points() if self.is_splitter_active else None
        self.diffuser.get_blade_from_control_points() if self.is_diffuser_active else None
        self.channel.get_channel_from_control_points() if self.is_channel_active else None

    def save(self, path: str):
        _, file_extension = os.path.splitext(path)
        if file_extension != ".bin":
            path += ".bin"

        with open(path, "wb") as f:
            joblib.dump(self, f)

        print(f"{self.__class__.__name__} saved to {path}")

    @staticmethod
    def load(path: str):
        if not os.path.exists(path):
            raise ValueError(f"{path} does not exist")
        return joblib.load(path)



import re
from collections import OrderedDict
from copy import copy, deepcopy

import numpy as np

from pynumeca.reader.iecEntry import iecEntry
from pynumeca.reader.iecGroup import iecGroup
from pynumeca.reader.niBladeGeometryEntry import niBladeGeometryEntry
from pynumeca.reader.zrCurveEntry import zrCurveEntry


class numecaParser(OrderedDict):
    def __init__(self, filename=""):
        super().__init__()
        self.fileName = filename
        self.stringData = ""

        if filename != "":
            self.load(filename)
            self.parsing()

    def load(self, filename):
        self.fileName = filename
        try:
            with open(self.fileName, "r") as file:
                self.stringData = file.readlines()
        except Exception as e:
            self.error_notification(e)
        self.parsing()

    def extract_d3d_parameters_string(self):
        output_string = ""
        try:
            number_of_parameters = self["ROOT"]["DESIGN3D_COMPUTATION_0"][
                "DATABASE_GENERATION_MINAMO_0"
            ]["SIMULATION_TASK_MANAGER_0"]["SIMULATION_TASKS_0"][
                "SIMULATION_TASK_PARAM_GEOM_MODEL_0"
            ][
                "PARAMETRIC_LAYER_0"
            ][
                "PARAMETERS_0"
            ][
                "NUMBER_OF_PARAMETERS"
            ].value

            output_string += number_of_parameters + "\n"

            for key, value in self["ROOT"]["DESIGN3D_COMPUTATION_0"][
                "DATABASE_GENERATION_MINAMO_0"
            ]["SIMULATION_TASK_MANAGER_0"]["SIMULATION_TASKS_0"][
                "SIMULATION_TASK_PARAM_GEOM_MODEL_0"
            ][
                "PARAMETRIC_LAYER_0"
            ][
                "PARAMETERS_0"
            ].items():
                if "PARAMETER_" in key:
                    output_string += value["NAME"].value + "\n"

            return output_string

        except Exception as e:
            raise CustomError

    def extract_d3d_parameters_list(self):
        output_list = []
        try:
            for key, value in self["ROOT"]["DESIGN3D_COMPUTATION_0"][
                "DATABASE_GENERATION_MINAMO_0"
            ]["SIMULATION_TASK_MANAGER_0"]["SIMULATION_TASKS_0"][
                "SIMULATION_TASK_PARAM_GEOM_MODEL_0"
            ][
                "PARAMETRIC_LAYER_0"
            ][
                "PARAMETERS_0"
            ].items():
                if "PARAMETER_" in key:
                    output_list.append(value["NAME"].value)

            return output_list
        except Exception as e:
            raise CustomError

    def create_d3d_default_parameter(self):
        stringlist = []
        stringlist.append("                     NI_BEGIN\tPARAMETER\n")
        stringlist.append(
            "                        NAME                           DEFAULT\n"
        )
        stringlist.append(
            "                        PARAMETRIC_TYPE                DOUBLE\n"
        )
        stringlist.append(
            "                        LIMIT_MIN                      -1000000000\n"
        )
        stringlist.append(
            "                        LIMIT_MAX                      1000000000\n"
        )
        stringlist.append(
            "                        VALUE                          0.002\n"
        )
        stringlist.append(
            "                        VALUE_MIN                      0.001\n"
        )
        stringlist.append(
            "                        VALUE_MAX                      0.003\n"
        )
        stringlist.append("                        VALUE_REF                      1\n")
        stringlist.append("                        NB_LEVELS                      2\n")
        stringlist.append(
            "                        QUANTITY_TYPE                  VALUE\n"
        )
        stringlist.append(
            "                        UNCERTAIN                      FALSE\n"
        )
        stringlist.append(
            "                        REDUCTION                      FALSE\n"
        )
        stringlist.append("                     NI_END\tPARAMETER\n")
        return self.deepTree(stringlist, 0)

    def create_d3d_parameter(self, name, type, value, min=None, max=None):
        new_parameter, _ = self.create_d3d_default_parameter()
        new_parameter["NAME"].value = name
        new_parameter["PARAMETRIC_TYPE"].value = type
        new_parameter["VALUE"].value = value
        if min is not None and max is not None:
            new_parameter["VALUE_MIN"].value = min
            new_parameter["VALUE_MAX"].value = max
        else:
            new_parameter["VALUE_MIN"].value = value
            new_parameter["VALUE_MAX"].value = value
        return new_parameter

    def add_d3d_parameter(self, *arg):
        parameters_layer = self["ROOT"]["DESIGN3D_COMPUTATION_0"][
            "DATABASE_GENERATION_MINAMO_0"
        ]["SIMULATION_TASK_MANAGER_0"]["SIMULATION_TASKS_0"][
            "SIMULATION_TASK_PARAM_GEOM_MODEL_0"
        ][
            "PARAMETRIC_LAYER_0"
        ][
            "PARAMETERS_0"
        ]

        if len(arg) == 1 and isinstance(arg[0], iecGroup):
            new_parameter = arg[0]
        elif len(arg) == 3:
            new_parameter = self.create_d3d_parameter(arg[0], arg[1], arg[2])
        elif len(arg) == 5:
            new_parameter = self.create_d3d_parameter(
                arg[0], arg[1], arg[2], arg[3], arg[4]
            )
        else:
            raise CustomError

        number_of_parameters = parameters_layer["NUMBER_OF_PARAMETERS"].value
        new_number_of_parameters = str(int(number_of_parameters) + 1)
        parameters_layer["NUMBER_OF_PARAMETERS"].value = new_number_of_parameters

        n = 0
        new_group_name = "PARAMETER" + "_" + str(n)
        while new_group_name in parameters_layer:
            n += 1
            new_group_name = "PARAMETER" + "_" + str(n)

        self["ROOT"]["DESIGN3D_COMPUTATION_0"]["DATABASE_GENERATION_MINAMO_0"][
            "SIMULATION_TASK_MANAGER_0"
        ]["SIMULATION_TASKS_0"]["SIMULATION_TASK_PARAM_GEOM_MODEL_0"][
            "PARAMETRIC_LAYER_0"
        ][
            "PARAMETERS_0"
        ] = self.insert_key_value(
            parameters_layer, new_group_name, "NI_END_PARAMETERS", new_parameter
        )

    def remove_d3d_parameter(self, number=None):
        parameters_layer = self["ROOT"]["DESIGN3D_COMPUTATION_0"][
            "DATABASE_GENERATION_MINAMO_0"
        ]["SIMULATION_TASK_MANAGER_0"]["SIMULATION_TASKS_0"][
            "SIMULATION_TASK_PARAM_GEOM_MODEL_0"
        ][
            "PARAMETRIC_LAYER_0"
        ][
            "PARAMETERS_0"
        ]
        number_of_parameters = parameters_layer["NUMBER_OF_PARAMETERS"].value

        if int(number_of_parameters) > 0 and number is None:
            new_number_of_parameters = str(int(number_of_parameters) - 1)
            parameters_layer["NUMBER_OF_PARAMETERS"].value = new_number_of_parameters

            n = 0
            last_group_name = "PARAMETER" + "_" + str(n)
            while last_group_name in parameters_layer:
                n += 1
                last_group_name = "PARAMETER" + "_" + str(n)
            n -= 1
            last_group_name = "PARAMETER" + "_" + str(n)

            self["ROOT"]["DESIGN3D_COMPUTATION_0"]["DATABASE_GENERATION_MINAMO_0"][
                "SIMULATION_TASK_MANAGER_0"
            ]["SIMULATION_TASKS_0"]["SIMULATION_TASK_PARAM_GEOM_MODEL_0"][
                "PARAMETRIC_LAYER_0"
            ][
                "PARAMETERS_0"
            ].pop(
                last_group_name
            )

        elif int(number_of_parameters) > 0 and (
            isinstance(number, int) or isinstance(number, str)
        ):
            new_number_of_parameters = str(int(number_of_parameters) - 1)
            parameters_layer["NUMBER_OF_PARAMETERS"].value = new_number_of_parameters

            if str(number).isdigit():
                last_group_name = "PARAMETER" + "_" + str(number)
            elif number in self.extract_d3d_parameters_list():
                last_group_name = ""
                for key, value in parameters_layer.items():
                    if "PARAMETER_" in key and number == value["NAME"].value:
                        last_group_name = key
            else:
                raise CustomError

            self["ROOT"]["DESIGN3D_COMPUTATION_0"]["DATABASE_GENERATION_MINAMO_0"][
                "SIMULATION_TASK_MANAGER_0"
            ]["SIMULATION_TASKS_0"]["SIMULATION_TASK_PARAM_GEOM_MODEL_0"][
                "PARAMETRIC_LAYER_0"
            ][
                "PARAMETERS_0"
            ].pop(
                last_group_name
            )

        else:
            pass

    @staticmethod
    def insert_key_value(a_dict, key, pos_key, value):
        new_dict = iecGroup()
        for k, v in a_dict.items():
            if k == pos_key:
                new_dict[key] = value  # insert new key
            new_dict[k] = v
        return new_dict

    def parsing(self):
        if isinstance(self.stringData, list) and "GEOMETRY TURBO" in self.stringData[0]:
            self.stringData[0] = "GEOMETRY-TURBO\n"
        self["ROOT"], temp = self.deepTree(self.stringData, 0)
        self["ROOT"] = self.deep_reorder_zrcurve(self["ROOT"])
        # self.merge_ZR()
        self.cleanUnusedBasicCurves()

    def deep_reorder_zrcurve(self, group):
        new_group = iecGroup()
        for key, value in group.items():
            if isinstance(value, iecGroup) and "zrcurve" in (list(value.items()))[0][0]:
                new_group[key] = zrCurveEntry(value)
            elif isinstance(value, iecGroup) and any(
                x in (list(value.items()))[0][0]
                for x in ["nibladegeometry", "NIBLADEGEOMETRY"]
            ):
                new_group[key] = niBladeGeometryEntry(value)
                # print (key)
                # new_group[key] = self.deepReorderZRcurve(value)
            #    new_group[key] = niBladeGeometryEntry(value)
            elif isinstance(value, iecGroup):
                new_group[key] = self.deep_reorder_zrcurve(value)
            else:
                new_group[key] = value
        return new_group

    def deepTree(self, string_data, line_index):
        newcall = True
        group = iecGroup()
        while line_index < len(string_data):
            line_string = string_data[line_index]
            entry = iecEntry()
            entry.parseSimpleEntry(line_string)
            if newcall and "NI_BEGIN" in line_string:
                # key = entry.key + "_" + entry.value
                if isinstance(entry.value, list):
                    key = entry.key + "_" + entry.tag + "_" + entry.value[0]
                else:
                    key = entry.key + "_" + entry.value
                group[key] = entry
                newcall = False
                line_index += 1
            elif line_index == 0 and entry.key == "GEOMETRY-TURBO":
                string_data[0] = "GEOMETRY TURBO\n"
                group_name = "GEOMTURBO"
                group[group_name], line_index = self.deepTree(string_data, line_index)
            elif line_index == 0 and entry.value == "TURBO":
                key = entry.key + "_" + entry.value
                group[key] = entry
                newcall = False
                line_index += 1
            elif not newcall and "NI_BEGIN" in line_string:
                n = 0
                if isinstance(entry.value, list):
                    group_name = entry.tag + "_" + entry.value[0] + "_" + str(n)
                else:
                    group_name = entry.value + "_" + str(n)
                while group_name in group:
                    n += 1
                    if isinstance(entry.value, list):
                        group_name = entry.tag + "_" + entry.value[0] + "_" + str(n)
                    else:
                        group_name = entry.value + "_" + str(n)

                group[group_name], line_index = self.deepTree(string_data, line_index)
            elif "NI_END" in line_string:
                key = entry.key + "_" + entry.value
                group[key] = entry
                newcall = False
                line_index += 1
                return group, line_index
            elif line_string.lstrip(" ").startswith(("*", "#")):
                n = 0
                key = "comment" + "_" + str(n)
                while key in group:
                    n += 1
                    key = "comment_" + str(n)
                group[key] = entry
                newcall = False
                line_index += 1
            elif line_string == "\n":
                n = 0
                key = "empty_line" + "_" + str(n)
                while key in group:
                    n += 1
                    key = "empty_line_" + str(n)
                group[key] = entry
                newcall = False
                line_index += 1
            else:
                key = entry.key
                n = 0
                while key in group:
                    n += 1
                    key = entry.key + "_" + str(n)
                group[key] = entry
                newcall = False
                line_index += 1

        return group, line_index

    @staticmethod
    def error_notification(e):
        print(e)
        print("Reading Error Notification")

    def outputString(self):
        return self["ROOT"].outputString()

    def checkSetup(self):
        simulation_task_manager = self["ROOT"]["DESIGN3D_COMPUTATION_0"][
            "DATABASE_GENERATION_MINAMO_0"
        ]["SIMULATION_TASK_MANAGER_0"]

        simulation_task_manager["MODEL_EXTENSION"].value = "prt"
        simulation_task_manager["CFD_GEOM_EXTENSION"].value = "stp"

        for idx, group in enumerate(
            (
                simulation_task_manager["SIMULATION_TASKS_0"][
                    "SIMULATION_TASK_PARAM_GEOM_MODEL_0"
                ],
                simulation_task_manager["SIMULATION_TASKS_0"][
                    "SIMULATION_TASK_IGG_AUTOGRID_0"
                ],
                simulation_task_manager["SIMULATION_TASKS_0"][
                    "SIMULATION_TASK_EURANUS_TURBO_0"
                ],
                simulation_task_manager["SIMULATION_TASKS_0"][
                    "SIMULATION_TASK_CFVIEW_0"
                ],
            )
        ):
            group["TASK_PROCESSING"].value = str(
                int(not (idx % 3))
            )  # Activate only n.0 and n.3
            group["CHAIN_TYPE"].value = "EXTERNAL"
            group["IMPORT_SCRIPTS"].value = "1"

    def setParamScript(self, script_name):
        self["ROOT"]["DESIGN3D_COMPUTATION_0"]["DATABASE_GENERATION_MINAMO_0"][
            "SIMULATION_TASK_MANAGER_0"
        ]["SIMULATION_TASKS_0"]["SIMULATION_TASK_PARAM_GEOM_MODEL_0"][
            "EXTERNAL_SCRIPT"
        ].value = (
            '"' + script_name + '"'
        )

    def setMeshScript(self, script_name):
        self["ROOT"]["DESIGN3D_COMPUTATION_0"]["DATABASE_GENERATION_MINAMO_0"][
            "SIMULATION_TASK_MANAGER_0"
        ]["SIMULATION_TASKS_0"]["SIMULATION_TASK_IGG_AUTOGRID_0"][
            "EXTERNAL_SCRIPT"
        ].value = (
            '"' + script_name + '"'
        )

    def setSolverScript(self, script_name):
        self["ROOT"]["DESIGN3D_COMPUTATION_0"]["DATABASE_GENERATION_MINAMO_0"][
            "SIMULATION_TASK_MANAGER_0"
        ]["SIMULATION_TASKS_0"]["SIMULATION_TASK_EURANUS_TURBO_0"][
            "EXTERNAL_SCRIPT"
        ].value = (
            '"' + script_name + '"'
        )

    def setPostProScript(self, script_name):
        self["ROOT"]["DESIGN3D_COMPUTATION_0"]["DATABASE_GENERATION_MINAMO_0"][
            "SIMULATION_TASK_MANAGER_0"
        ]["SIMULATION_TASKS_0"]["SIMULATION_TASK_CFVIEW_0"]["EXTERNAL_SCRIPT"].value = (
            '"' + script_name + '"'
        )

    def createPostProDefaultMacro(self):
        stringlist = []
        stringlist.append("                  NI_BEGIN\tUSER_DEFINED_MACRO\n")
        stringlist.append(
            "                     SOLUTION_FILE                                      3D\n"
        )
        stringlist.append(
            "                     MACRO_FILE                                         myPath/myScript.py\n"
        )
        stringlist.append(
            "                     DATA_FILE                                          myScript.dat\n"
        )
        stringlist.append(
            "                     QUANTITY_NAME                                      myQuantity\n"
        )
        stringlist.append(
            "                     DIM                                                SCALAR\n"
        )
        stringlist.append(
            "                     ACTIVE                                             1\n"
        )
        stringlist.append("                  NI_END\tUSER_DEFINED_MACRO\n")
        return self.deepTree(stringlist, 0)

    def createPostProMacro(self, script_name):
        new_post_pro_macro, _ = self.createPostProDefaultMacro()
        new_post_pro_macro["MACRO_FILE"].value = "macro_CFV/" + script_name
        new_post_pro_macro["DATA_FILE"].value = script_name.rsplit(".", 1)[0] + ".dat"
        new_post_pro_macro["QUANTITY_NAME"].value = script_name.rsplit(".", 1)[0]
        return new_post_pro_macro

    def addPostProMacro(self, script_name):
        postProMacroLayer = self["ROOT"]["DESIGN3D_COMPUTATION_0"][
            "DATABASE_GENERATION_MINAMO_0"
        ]["SIMULATION_TASK_MANAGER_0"]["SIMULATION_TASKS_0"][
            "SIMULATION_TASK_CFVIEW_0"
        ][
            "USER_DEFINED_MACROS_EXTERNAL_0"
        ]

        if isinstance(script_name, iecGroup):
            new_post_pro_macro = script_name
        elif isinstance(script_name, str):
            new_post_pro_macro = self.createPostProMacro(script_name)
        else:
            raise CustomError

        n = 0
        new_group_name = "USER_DEFINED_MACRO" + "_" + str(n)
        while new_group_name in postProMacroLayer:
            n += 1
            new_group_name = "USER_DEFINED_MACRO" + "_" + str(n)

        self["ROOT"]["DESIGN3D_COMPUTATION_0"]["DATABASE_GENERATION_MINAMO_0"][
            "SIMULATION_TASK_MANAGER_0"
        ]["SIMULATION_TASKS_0"]["SIMULATION_TASK_CFVIEW_0"][
            "USER_DEFINED_MACROS_EXTERNAL_0"
        ] = self.insert_key_value(
            postProMacroLayer,
            new_group_name,
            "NI_END_USER_DEFINED_MACROS_EXTERNAL",
            new_post_pro_macro,
        )

    def removePostProMacro(self, number=None):
        post_pro_macro_layer = self["ROOT"]["DESIGN3D_COMPUTATION_0"][
            "DATABASE_GENERATION_MINAMO_0"
        ]["SIMULATION_TASK_MANAGER_0"]["SIMULATION_TASKS_0"][
            "SIMULATION_TASK_CFVIEW_0"
        ][
            "USER_DEFINED_MACROS_EXTERNAL_0"
        ]

        if number is None:
            n = 0
            last_group_name = "USER_DEFINED_MACRO" + "_" + str(n)
            while last_group_name in post_pro_macro_layer:
                n += 1
                last_group_name = "USER_DEFINED_MACRO" + "_" + str(n)
            n -= 1
            last_group_name = "USER_DEFINED_MACRO" + "_" + str(n)

        elif isinstance(number, int):
            last_group_name = "USER_DEFINED_MACRO" + "_" + str(number)

        else:
            raise CustomError

        self["ROOT"]["DESIGN3D_COMPUTATION_0"]["DATABASE_GENERATION_MINAMO_0"][
            "SIMULATION_TASK_MANAGER_0"
        ]["SIMULATION_TASKS_0"]["SIMULATION_TASK_CFVIEW_0"][
            "USER_DEFINED_MACROS_EXTERNAL_0"
        ].pop(
            last_group_name
        )

    def exportNpyArray(self, row_number=0, blade_number=0):
        ni_blade_geometry = self.retrieveNiBladeGeometry(row_number, blade_number)

        suction = []
        pressure = []
        for item in ni_blade_geometry.suctionList:
            section = np.vstack(
                [item.X, item.Y, item.Z, np.zeros(item.numberOfPointsInt)]
            ).transpose()
            suction.append(section)
        for item in ni_blade_geometry.pressureList:
            section = np.vstack(
                [item.X, item.Y, item.Z, np.ones(item.numberOfPointsInt)]
            ).transpose()
            section = np.flip(section, axis=0)
            pressure.append(section)
        return np.array([(np.asarray(suction), np.asarray(pressure))], dtype=float)

    def exportNpyArrayCyl(self, row_number=0, blade_number=0):
        ni_blade_geometry = self.retrieveNiBladeGeometry(row_number, blade_number)
        if not self.cylCoordDefined(row_number, blade_number):
            self.convertCartesian2Cyl(row_number, blade_number)
        suction = []
        pressure = []
        for item in ni_blade_geometry.suctionList:
            section = np.vstack(
                [item.R, item.THETA, item.Z, np.zeros(item.numberOfPointsInt)]
            ).transpose()
            suction.append(section)
        for item in ni_blade_geometry.pressureList:
            section = np.vstack(
                [item.R, item.THETA, item.Z, np.ones(item.numberOfPointsInt)]
            ).transpose()
            section = np.flip(section, axis=0)
            pressure.append(section)
        return np.array([(np.asarray(suction), np.asarray(pressure))], dtype=float)

    def convertCartesian2Cyl(self, row_number=0, blade_number=0):
        ni_blade_geometry = self.retrieveNiBladeGeometry(row_number, blade_number)
        for item in ni_blade_geometry.suctionList:
            item.R = []
            item.THETA = []
            for index in range(len(item.X)):
                item.R.append(np.hypot(item.Y[index], item.X[index]))
                item.THETA.append(np.arctan2(item.Y[index], item.X[index]))
        for item in ni_blade_geometry.pressureList:
            item.R = []
            item.THETA = []
            for index in range(len(item.X)):
                item.R.append(np.hypot(item.Y[index], item.X[index]))
                item.THETA.append(np.arctan2(item.Y[index], item.X[index]))

    def cylCoordDefined(self, row_number=0, blade_number=0):
        ni_blade_geometry = self.retrieveNiBladeGeometry(row_number, blade_number)
        check = 0
        for item in ni_blade_geometry.suctionList:
            check += len(item.THETA) + len(item.R)
        for item in ni_blade_geometry.pressureList:
            check += len(item.THETA) + len(item.R)
        return bool(check)

    def importNpyArrayCyl(self, array, row_number=0, blade_number=0):
        # array shape (2,nSections,nPoints,4)
        ni_blade_geometry = self.retrieveNiBladeGeometry(row_number, blade_number)
        n_sections_from_array = array.shape[1]
        ni_blade_geometry.setNumberOfSections(n_sections_from_array)
        for idx, suction in enumerate(ni_blade_geometry.suctionList):
            new_list = array[0][idx].transpose().tolist()
            suction.updateArraysCyl(new_list[0], new_list[1], new_list[2])
        for idx, pressure in enumerate(ni_blade_geometry.pressureList):
            new_list = array[1][idx].transpose().tolist()
            pressure.updateArraysCyl(
                new_list[0][::-1], new_list[1][::-1], new_list[2][::-1]
            )

    def importNpyArray(self, array, row_number=0, blade_number=0):
        # array shape (2,nSections,nPoints,4)
        ni_blade_geometry = self.retrieveNiBladeGeometry(row_number, blade_number)
        n_sections_from_array = array.shape[1]
        ni_blade_geometry.setNumberOfSections(n_sections_from_array)
        for idx, suction in enumerate(ni_blade_geometry.suctionList):
            new_list = array[0][idx].transpose().tolist()
            suction.updateArrays(new_list[0], new_list[1], new_list[2])
        for idx, pressure in enumerate(ni_blade_geometry.pressureList):
            new_list = array[1][idx].transpose().tolist()
            pressure.updateArrays(
                new_list[0][::-1], new_list[1][::-1], new_list[2][::-1]
            )

    def retrieveNiBladeGeometry(self, row_number=0, blade_number=0):
        row_occurence = -1
        blade_occurence = -1
        blade_key = ""

        row_key = self.retrieve_row_key(row_number)

        for key in self["ROOT"]["GEOMTURBO"][row_key].keys():
            if "NIBlade" in key or "NIBLADE" in key:
                blade_occurence += 1
                if blade_occurence == blade_number:
                    blade_key = key

        if row_key == "":
            print("Row not defined")
        if blade_key == "":
            print("Blade not defined")

        if "NIROW" in row_key:
            ni_blade_geometry = self["ROOT"]["GEOMTURBO"][row_key][blade_key][
                "NIBLADEGEOMETRY_0"
            ]
        elif "nirow" in row_key:
            ni_blade_geometry = self["ROOT"]["GEOMTURBO"][row_key][blade_key][
                "nibladegeometry_0"
            ]
        else:
            print("ERROR")

        return ni_blade_geometry

    def set_row_periodicity(self, periodicity, rowNumber=0):
        row_key = self.retrieve_row_key(rowNumber)

        if row_key == "" or row_key == None:
            print("Row not defined")
        else:
            self["ROOT"]["GEOMTURBO"][row_key]["PERIODICITY"].value = str(periodicity)

    def get_row_periodicity(self, rowNumber=0):
        row_key = self.retrieve_row_key(rowNumber)
        if row_key == "" or row_key == None:
            print("Row not defined")
        else:
            return self["ROOT"]["GEOMTURBO"][row_key]["PERIODICITY"].value

    def set_row_speed(self, row_speed, rowNumber=0):
        row_key = self.retrieve_row_key(rowNumber)

        if row_key == "" or row_key == None:
            print("Row not defined")
        else:
            try:
                self["ROOT"]["GEOMTURBO"][row_key]["ROTATION_SPEED"].value = str(
                    row_speed
                )
            except:
                rot_speed = iecEntry()
                rot_speed.key = "ROTATION_SPEED"
                rot_speed.value = str(row_speed)
                rot_speed.leadingSpaceString = self["ROOT"]["GEOMTURBO"][row_key][
                    "PERIODICITY"
                ].leadingSpaceString
                rot_speed.keyTagSpaceString = self["ROOT"]["GEOMTURBO"][row_key][
                    "PERIODICITY"
                ].keyTagSpaceString
                rot_speed.tagValueSpaceString = self["ROOT"]["GEOMTURBO"][row_key][
                    "PERIODICITY"
                ].tagValueSpaceString
                rot_speed.trailingSpaceString = self["ROOT"]["GEOMTURBO"][row_key][
                    "PERIODICITY"
                ].trailingSpaceString
                last_item_key = list(self["ROOT"]["GEOMTURBO"][row_key].keys())[-1]
                last_item_copy = deepcopy(
                    self["ROOT"]["GEOMTURBO"][row_key][last_item_key]
                )
                del self["ROOT"]["GEOMTURBO"][row_key][last_item_key]
                self["ROOT"]["GEOMTURBO"][row_key]["ROTATION_SPEED"] = rot_speed
                self["ROOT"]["GEOMTURBO"][row_key][last_item_key] = last_item_copy

    def get_row_speed(self, rowNumber=0):
        row_key = self.retrieve_row_key(rowNumber)
        if row_key == "" or row_key == None:
            print("Row not defined")
        else:
            try:
                row_speed = self["ROOT"]["GEOMTURBO"][row_key]["ROTATION_SPEED"].value
                return row_speed
            except:
                return "NaN"

    def retrieve_row_key(self, rowNumber):
        row_occurence = -1
        for key in self["ROOT"]["GEOMTURBO"].keys():
            if "nirow" in key or "NIROW" in key:
                row_occurence += 1
                if row_occurence == rowNumber:
                    return key
        return None

    def merge_ZR(self):
        if "GEOMTURBO" not in self["ROOT"].keys():
            return
        basic_curve_dict = self.get_basic_curve_dict()
        hub_vertex_list = self.get_vertex_list("channel_curve_hub_0")
        shroud_vertex_list = self.get_vertex_list("channel_curve_shroud_0")

        self.append_and_update_curves(basic_curve_dict, hub_vertex_list)
        self.append_and_update_curves(basic_curve_dict, shroud_vertex_list)

    def get_vertex_list(self, channel_name):
        vertex_list = []
        for key in self["ROOT"]["GEOMTURBO"]["CHANNEL_0"][channel_name].keys():
            if "VERTEX" in key:
                vertex_list.append(
                    (
                        channel_name,
                        key,
                        self["ROOT"]["GEOMTURBO"]["CHANNEL_0"][channel_name][key].value[
                            0
                        ],
                    )
                )
        return vertex_list

    def get_basic_curve_dict(self):
        basic_curve_dict = {}
        for key in self["ROOT"]["GEOMTURBO"]["CHANNEL_0"].keys():
            if "basic_curve" in key:
                try:
                    name = self["ROOT"]["GEOMTURBO"]["CHANNEL_0"][key]["NAME"].value
                    zrList = self["ROOT"]["GEOMTURBO"]["CHANNEL_0"][key]["zrcurve_0"]
                    basic_curve_dict[name] = (key, zrList)
                except:
                    pass
        return basic_curve_dict

    def append_and_update_curves(self, basic_curve_dict, vertex_list):
        for item in basic_curve_dict.items():
            curve = item[1][1]
            if curve.numberOfPoints == 2:
                # QUI bisogna aggiungere i punti
                # pass
                curve.uniformIncreasePointsNumber(50)

        base_curve = vertex_list[1][2]
        for curve in vertex_list[2:]:
            basic_curve_dict[base_curve][1].append(basic_curve_dict[curve[2]][1])
            key_to_remove = basic_curve_dict[curve[2]][0]
            del self["ROOT"]["GEOMTURBO"]["CHANNEL_0"][key_to_remove]
            del self["ROOT"]["GEOMTURBO"]["CHANNEL_0"][curve[0]][curve[1]]

    def extractChannelCurves(self, channel_curve_name):
        basic_curve_dict = self.get_basic_curve_dict()
        # channel_curve_name possible values = channel_curve_hub_0
        #                                      channel_curve_shroud_0
        if "shroud" in channel_curve_name:
            zeros_or_ones = np.ones
        elif "hub" in channel_curve_name:
            zeros_or_ones = np.zeros
        else:
            raise Exception("ERROR ZERO ONE HUB SHROUD")

        curve_array_list = []
        avoid_first_occurrence = True
        for key, item in self["ROOT"]["GEOMTURBO"]["CHANNEL_0"][
            channel_curve_name
        ].items():
            if "VERTEX" in key:
                if not avoid_first_occurrence:
                    curve_name = item.value[0]
                    curve = basic_curve_dict[curve_name][1]
                    curve_array = np.vstack(
                        [
                            np.zeros(curve.numberOfPoints),
                            curve.R,
                            curve.Z,
                            zeros_or_ones(curve.numberOfPoints),
                        ]
                    ).transpose()
                    curve_array = np.expand_dims(curve_array, axis=0)
                    curve_array_list.append(curve_array)
                else:
                    avoid_first_occurrence = False
        return curve_array_list

    def exportZRNpyArraysList(self):
        hub_array_list = self.extractChannelCurves("channel_curve_hub_0")
        shroud_array_list = self.extractChannelCurves("channel_curve_shroud_0")
        return (hub_array_list, shroud_array_list)

    def vStackArray(self, arrayList):
        if arrayList[0].shape[1] == 2:
            vStackArray = self.fillZRArray(arrayList[0], 50)
        else:
            vStackArray = arrayList[0]
        for array in arrayList[1:]:
            if array.shape[1] == 2:
                array = self.fillZRArray(array, 50)
            vStackArray = np.append(vStackArray, array[:, 1:, :], axis=1)
        return vStackArray

    def fillZRArray(self, array, nPoints):
        if array[0, 0, 3] == 0:
            zeros_or_ones = np.zeros
        elif array[0, 0, 3] == 1:
            zeros_or_ones = np.ones
        else:
            raise Exception("ERROR ZERO ONE HUB SHROUD")
        ref = np.array([0, 1])
        X = np.linspace(0, 1, num=nPoints)
        R = np.interp(X, ref, array[0, :, 1])
        Z = np.interp(X, ref, array[0, :, 2])
        newArray = np.vstack(
            [np.zeros(nPoints), R, Z, zeros_or_ones(nPoints)]
        ).transpose()
        return np.expand_dims(newArray, axis=0)

    def exportZRNpyArrays(self):
        hub_array_list, shroud_array_list = self.exportZRNpyArraysList()
        return (self.vStackArray(hub_array_list), self.vStackArray(shroud_array_list))

    def importZRNpyArrays(self, hub_section, shroud_section):
        self.importZRNpyHubArray(hub_section)
        self.importZRNpyShroudArray(shroud_section)
        pass

    def importZRNpyHubArray(self, hub_section):
        self.importZRNpyGenericArrayWithKey(hub_section, "channel_curve_hub_0")
        pass

    def importZRNpyShroudArray(self, shroud_section):
        self.importZRNpyGenericArrayWithKey(shroud_section, "channel_curve_shroud_0")
        pass

    def importZRNpyGenericArrayWithKey(self, section, key):
        basic_curve_dict = self.get_basic_curve_dict()
        curve_name = self["ROOT"]["GEOMTURBO"]["CHANNEL_0"][key]["VERTEX"].value[0]
        curve = basic_curve_dict[curve_name][1]
        curve.updateArrays(section[:, 2], section[:, 1])
        curve.numberOfPoints = section.shape[0]

    def importZRNpyArray(self, section):
        if np.all(section[0, :, 3] == 0):
            self.setNumberOfBasicCurvesHub(1)
            self.importZRNpyHubArray(section[0])
        elif np.all(section[0, :, 3] == 1):
            self.setNumberOfBasicCurvesShroud(1)
            self.importZRNpyShroudArray(section[0])
        else:
            raise Exception("ERROR: NON COHERENT ARRAY")

    def importZRNpyArrayList(self, sections_list):
        if np.all(sections_list[0][0, :, 3] == 0):
            # Hub section
            self.setNumberOfBasicCurvesHub(len(sections_list))
            self.updateGenericSectionsFromList(sections_list, "channel_curve_hub_0")
        elif np.all(sections_list[0][0, :, 3] == 1):
            # Shroud section
            self.setNumberOfBasicCurvesShroud(len(sections_list))
            self.updateGenericSectionsFromList(sections_list, "channel_curve_shroud_0")

    def updateGenericSectionsFromList(self, section_list, hub_or_shroud):
        names_of_basic_curve_list = self.getNamesListOfBasicCurvesGeneric(hub_or_shroud)
        if len(names_of_basic_curve_list) != len(section_list):
            raise Exception("ERROR: mismatch between endwalls sections number")

        for idx, curve_name in enumerate(names_of_basic_curve_list):
            basic_curve_dict = self.get_basic_curve_dict()
            curve = basic_curve_dict[curve_name][1]
            curve.updateArrays(section_list[idx][0, :, 2], section_list[idx][0, :, 1])
            curve.numberOfPoints = section_list[idx].shape[1]

    def importZRNpyArray2(self, section):
        if isinstance(section, list) and isinstance(section[0], np.ndarray):
            # Sections divided in list
            self.importZRNpyArrayList(section)
            pass
        elif isinstance(section, np.ndarray):
            # Single section
            self.importZRNpyArray(section)
        else:
            raise Exception("ERROR: wrong format of endwalls ")

    def getNumberOfBasicCurvesHub(self):
        return len(self.getNamesListOfBasicCurvesHub())

    def getNumberOfBasicCurvesShroud(self):
        return len(self.getNamesListOfBasicCurvesShroud())

    def getNamesListOfBasicCurvesHub(self):
        return self.getNamesListOfBasicCurvesGeneric("channel_curve_hub_0")

    def getNamesListOfBasicCurvesShroud(self):
        return self.getNamesListOfBasicCurvesGeneric("channel_curve_shroud_0")

    def getNamesListOfBasicCurvesGeneric(self, hub_or_shroud):
        vertex_list = self["ROOT"]["GEOMTURBO"]["CHANNEL_0"][hub_or_shroud]
        vertex_number = -1  # Because the first one is always used twice
        names_list_of_basic_curves = []
        for key, item in vertex_list.items():
            if "VERTEX" in key:
                vertex_number += 1
                names_list_of_basic_curves.append(item.value[0])
        del names_list_of_basic_curves[0]
        return names_list_of_basic_curves

    def setNumberOfBasicCurvesHub(self, number):
        return self.setNumberOfBasicCurvesGeneric(number, "channel_curve_hub_0")
        pass

    def setNumberOfBasicCurvesShroud(self, number):
        return self.setNumberOfBasicCurvesGeneric(number, "channel_curve_shroud_0")
        pass

    def setNumberOfBasicCurvesGeneric(self, number, hub_or_shroud):
        basic_curve_dict = self.get_basic_curve_dict()
        names_list = self.getNamesListOfBasicCurvesGeneric(hub_or_shroud)
        if number < len(names_list):
            for i in range(len(names_list) - number):
                self.removeLastBasicCurveGeneric(hub_or_shroud)
        elif number > len(names_list):
            for i in range(number - len(names_list)):
                self.addBasicCurveGeneric(hub_or_shroud)

    def removeLastBasicCurveHub(self):
        return self.removeLastBasicCurveGeneric("channel_curve_hub_0")

    def removeLastBasicCurveShroud(self):
        return self.removeLastBasicCurveGeneric("channel_curve_shroud_0")

    def removeLastBasicCurveGeneric(self, hub_or_shroud):
        names_of_basic_curve_list = self.getNamesListOfBasicCurvesGeneric(hub_or_shroud)
        if len(names_of_basic_curve_list) <= 1:
            raise Exception("Minimum basic curve number already reached")
        basic_curve_dict = self.get_basic_curve_dict()
        vertex_list = self["ROOT"]["GEOMTURBO"]["CHANNEL_0"][hub_or_shroud]
        last_item_name = names_of_basic_curve_list[-1]
        for key, item in vertex_list.items():
            if item.value[0] == last_item_name:
                del vertex_list[key]
                break
        basic_curve_name_to_be_removed = basic_curve_dict[last_item_name][0]
        del self["ROOT"]["GEOMTURBO"]["CHANNEL_0"][basic_curve_name_to_be_removed]

    def addBasicCurveItem(self, new_item_name, new_curve_name):
        main_channel = self["ROOT"]["GEOMTURBO"]["CHANNEL_0"]
        first_item_key = next(iter(main_channel))

        for key, item in main_channel.items():
            if "basic_curve" in key:
                try:
                    zrList = self["ROOT"]["GEOMTURBO"]["CHANNEL_0"][key]["zrcurve_0"]
                    first_basic_curve = item
                    break
                except:
                    pass

        main_channel[new_item_name] = deepcopy(first_basic_curve)
        main_channel.move_to_end(new_item_name, False)
        main_channel.move_to_end(first_item_key, False)
        main_channel[new_item_name]["NAME"].value = new_curve_name

    def addBasicCurveHub(self):
        return self.addBasicCurveGeneric("channel_curve_hub_0")

    def addBasicCurveShroud(self):
        return self.addBasicCurveGeneric("channel_curve_shroud_0")

    def cleanUnusedBasicCurves(self):
        names_of_used_basic_curve_list = (
            self.getNamesListOfBasicCurvesHub() + self.getNamesListOfBasicCurvesShroud()
        )
        main_channel = self["ROOT"]["GEOMTURBO"]["CHANNEL_0"]
        list_of_keys_to_be_removed = []
        for key, item in main_channel.items():
            if (
                "basic_curve" in key
                and item["NAME"].value not in names_of_used_basic_curve_list
            ):
                list_of_keys_to_be_removed.append(key)

        for key in list_of_keys_to_be_removed:
            del main_channel[key]

    def addBasicCurveGeneric(self, hub_or_shroud):
        names_of_basic_curve_list = (
            self.getNamesListOfBasicCurvesHub() + self.getNamesListOfBasicCurvesShroud()
        )
        vertex_list = self["ROOT"]["GEOMTURBO"]["CHANNEL_0"][hub_or_shroud]
        last_curve_name = names_of_basic_curve_list[-1]
        last_vertex_name = list(vertex_list.keys())[-2]
        last_item_name = list(vertex_list.keys())[-1]
        last_number = re.search(r"\d+$", last_curve_name)

        if last_number:
            residual = re.search(r".+?(?=\d+$)", last_curve_name).group(0)
            last_number = int(last_number.group(0))
        else:
            residual = last_curve_name
            last_number = 0
        free_name_found = True
        counter = 0
        while free_name_found:
            counter += 1
            new_curve_name = residual + str(last_number + counter)
            free_name_found = new_curve_name in names_of_basic_curve_list
        free_vertex_key_found = True
        counter = 0
        while free_vertex_key_found:
            counter += 1
            new_vertex_key = "VERTEX_" + str(counter)
            free_vertex_key_found = new_vertex_key in vertex_list.keys()

        new_item = iecEntry(vertex_list[last_vertex_name])
        new_item.value = [new_curve_name, str(1)]
        vertex_list[new_vertex_key] = new_item
        vertex_list.move_to_end(last_item_name)

        basic_curve_dict = self.get_basic_curve_dict()

        free_name_found = True
        counter = 0
        while free_name_found:
            counter += 1
            new_item_name = "basic_curve_" + str(counter)
            free_name_found = {
                k: v for k, v in basic_curve_dict.items() if v[0] == new_item_name
            }

        self.addBasicCurveItem(new_item_name, new_curve_name)
        pass


class CustomError(Exception):
    pass



from __future__ import annotations

import os
import re
import time
from typing import List, Tuple

import numpy as np
import plotly.graph_objects as go

from pynumeca.reader.numecaParser import numecaParser


def find_between(text: str, start_keyword: str, end_keyword: str) -> list:
    return re.findall(f"(?<={start_keyword}).*?(?={end_keyword})", text, re.DOTALL)


class ArrayWithName(object):
    def __init__(self, name: str, array: np.ndarray | float):
        self.name = name
        self.array = array

    @staticmethod
    def get_curve_by_name(my_list: List["ArrayWithName"], names: List):
        for item in my_list:
            if item.name in names:
                return item


class GeomTurboParser(object):
    __impeller_wheel_name = ["Impeller_wheel", "Rotor_wheel", "row1", "impeller_wheel"]
    __cascade_wheel_name = ["Cascade_Diffuser_wheel", "Nozzle_wheel"]
    __main_blade_name = [
        "Impeller_main",
        "Rotor_main",
        "mainblade",
        "impeller_main",
        "MainBlade",
    ]
    __splitter_blade_name = ["Impeller_spl", "splitter", "splitter1"]
    __diffuser_blade_name = [
        "Cascade_Diffuser_main",
        "Nozzle_main",
        "diffuser",
        "diffuser1",
    ]
    __straight_lines_filling_points = 50

    def __init__(
        self,
        target_path: str,
        splitter_active: bool = True,
        diffuser_active: bool = True,
    ):
        if os.path.exists(target_path):
            with open(target_path) as f:
                self.raw_content = f.read()

        else:
            raise FileNotFoundError("File not found: %s" % target_path)

        self.splitter_active = splitter_active
        self.diffuser_active = diffuser_active

        (
            self.main_blade,
            self.splitter,
            self.diffuser,
            self.mb_periodicity,
            self.diff_periodicity,
        ) = self.get_blades()

        self.hub, self.shroud = self.get_channel()

    def plot_channel(self, show: bool = False) -> go.Figure:
        # Create scatter plot trace
        hub = go.Scatter(
            x=self.hub[0, :, 2],
            y=self.hub[0, :, 1],
            mode="markers+lines",
            name="Hub",
        )

        shroud = go.Scatter(
            x=self.shroud[0, :, 2],
            y=self.shroud[0, :, 1],
            mode="markers+lines",
            name="Shroud",
        )

        # Create layout
        layout = go.Layout(
            title="Channel", xaxis=dict(title="X-axis"), yaxis=dict(title="Y-axis")
        )

        # Create figure and add trace
        fig = go.Figure(data=[hub, shroud], layout=layout)

        # Display the plot
        fig.show() if show else None
        return fig

    @property
    def shape(self) -> Tuple[tuple, ...]:
        return self.main_blade.shape, self.splitter.shape, self.diffuser.shape

    def get_blades(self) -> Tuple[np.ndarray, ...]:
        self.raw_content = self.raw_content.replace(
            "NIBLADEGEOMETRY", "nibladegeometry"
        )

        blades_raw_content = find_between(
            self.raw_content, "NI_BEGIN NIBlade", "NI_END NIBlade"
        )

        if len(blades_raw_content) == 0:
            blades_raw_content = find_between(
                self.raw_content, "NI_BEGIN	NIBLADE", "NI_END	NIBLADE"
            )

            if len(blades_raw_content) == 0:
                raise ValueError("Cannot find any blade in geomTurbo file")

        wheels_raw_content = find_between(
            self.raw_content, "NI_BEGIN nirow", "NI_END nirow"
        )

        if len(wheels_raw_content) == 0:
            wheels_raw_content = find_between(
                self.raw_content, "NI_BEGIN	NIROW", "NI_END	NIROW"
            )

            if len(wheels_raw_content) == 0:
                raise ValueError("Cannot find any wheel in geomTurbo file")

        extracted_wheels = []

        for wheel in wheels_raw_content:
            name = (
                find_between(wheel, "NAME", "TYPE")[0]
                .replace(" ", "")
                .replace("\n", "")
            )

            name = name.replace("\t", "")

            try:
                periodicity = int(
                    find_between(wheel, "PERIODICITY", "ROTATION_SPEED")[0]
                    .replace(" ", "")
                    .replace("\n", "")
                )
            except:
                try:
                    periodicity = int(
                        find_between(wheel, "PERIODICITY", "\n")[0].replace("\t", "")
                    )
                except Exception as e:
                    print("cannot find row periodicity: ", e)

            extracted_wheels.append(ArrayWithName(name=name, array=periodicity))

        extracted_blades = []

        for blade in blades_raw_content:
            name = (
                find_between(blade, "NAME", "NI_BEGIN")[0]
                .replace(" ", "")
                .replace("\n", "")
            )

            if "\t" in name:
                name = name.replace("\t", "")

            blade_geom = find_between(
                blade, "NI_BEGIN nibladegeometry", "NI_END nibladegeometry"
            )

            if len(blade_geom) == 0:
                blade_geom = find_between(
                    blade, "NI_BEGIN\tnibladegeometry", "NI_END\tnibladegeometry"
                )

                if len(blade_geom) == 0:
                    raise ValueError("Cannot detect blade geometry")

            blade_geom = blade_geom[0]

            def extract_coordinates(text):
                lines = text.strip().split("\n")
                sections = []
                for i, line in enumerate(lines):
                    if "section" in line:
                        n_points = int(lines[i + 2].strip())
                        section = []
                        for j in range(n_points + 2):
                            actual_line = lines[i + j + 1]
                            if len(actual_line.strip().split()) == 3:
                                section.append(
                                    list(map(float, actual_line.strip().split()))
                                )

                        sections.append(section)
                return np.array(sections)

            suction_text = blade_geom.split("suction")[1].split("pressure")[0]
            pressure_text = blade_geom.split("suction")[1].split("pressure")[1]

            suction = extract_coordinates(suction_text)
            suction = np.concatenate(
                [suction, np.zeros(shape=(*suction.shape[:-1], 1))], axis=-1
            )
            pressure = extract_coordinates(pressure_text)
            pressure = np.flip(
                np.concatenate(
                    [pressure, np.ones(shape=(*pressure.shape[:-1], 1))], axis=-1
                ),
                axis=1,
            )

            extracted_blades.append(
                ArrayWithName(name=name, array=np.stack([suction, pressure], axis=0))
            )

        def ed(array: np.ndarray):
            return np.expand_dims(axis=0, a=array)

        main_blade = ed(
            ArrayWithName.get_curve_by_name(
                extracted_blades, self.__main_blade_name
            ).array
        )
        if self.splitter_active:
            splitter = ed(
                ArrayWithName.get_curve_by_name(
                    extracted_blades, self.__splitter_blade_name
                ).array
            )
        else:
            splitter = None

        if self.diffuser_active:
            diffuser = ed(
                ArrayWithName.get_curve_by_name(
                    extracted_blades, self.__diffuser_blade_name
                ).array
            )
        else:
            diffuser = None

        mb_periodicity = ArrayWithName.get_curve_by_name(
            extracted_wheels, self.__impeller_wheel_name
        ).array

        if self.diffuser_active:
            diff_periodicity = ArrayWithName.get_curve_by_name(
                extracted_wheels, self.__cascade_wheel_name
            ).array
        else:
            diff_periodicity = None

        return main_blade, splitter, diffuser, mb_periodicity, diff_periodicity

    def get_channel(self) -> Tuple[np.ndarray, ...]:
        channel_text = find_between(
            self.raw_content, "NI_BEGIN CHANNEL", "NI_END CHANNEL"
        )

        if len(channel_text) == 0:
            channel_text = find_between(
                self.raw_content, "NI_BEGIN\tCHANNEL", "NI_END\tCHANNEL"
            )

            if len(channel_text) == 0:
                raise ValueError("Cannot find channel in geomTurbo file")

        channel_text = channel_text[0]

        curves = find_between(
            channel_text, "NI_BEGIN basic_curve", "NI_END basic_curve"
        )

        if len(curves) == 0:
            curves = find_between(
                channel_text, "NI_BEGIN\tbasic_curve", "NI_END\tbasic_curve"
            )

        extracted_curves = []

        for curve in curves:
            name = (
                find_between(curve, "NAME", "DISCRETISATION")[0]
                .replace(" ", "")
                .replace("\n", "")
            )

            if "\t" in name:
                name = name.replace("\t", "")
            # Split the text into lines and remove the first line ("ZR")
            try:
                lines = (
                    find_between(curve, "ZR", "NI_END zrcurve")[0]
                    .strip()
                    .split("\n")[1:]
                )
            except:
                lines = (
                    find_between(curve, "ZR", "NI_END\tzrcurve")[0]
                    .strip()
                    .split("\n")[1:]
                )

            # Convert each line to a list of floats and stack them vertically to create a 2D numpy array
            array = np.vstack([np.fromstring(line, sep=" ") for line in lines])
            extracted_curves.append(ArrayWithName(name=name, array=array))

        hub_composition = find_between(
            channel_text, "NI_BEGIN channel_curve hub", "NI_END channel_curve hub"
        )

        if len(hub_composition) == 0:
            hub_composition = find_between(
                channel_text, "NI_BEGIN\tchannel_curve hub", "NI_END\tchannel_curve hub"
            )

            if len(hub_composition) == 0:
                raise ValueError("Cannot define hub curve composition")

        hub_composition = hub_composition[0]

        hub_curves = [
            f"curve_{item}"
            for item in [
                int(match.group(1))
                for match in re.finditer(r"curve_(\d+)", hub_composition)
            ]
        ]

        if len(hub_curves) == 0:
            hub_curves = [
                f"hub_crv_{item}"
                for item in [
                    int(match.group(1))
                    for match in re.finditer(r"hub_crv_(\d+)", hub_composition)
                ]
            ]

        seen = {}
        hub_curves = [seen.setdefault(x, x) for x in hub_curves if x not in seen]
        hub_arrays = [
            ArrayWithName.get_curve_by_name(extracted_curves, [item]).array
            for item in hub_curves
        ]
        hub_arrays = self.__fill_straight_curves(hub_arrays)
        hub_array = np.concatenate(hub_arrays)

        # hub_array, _ = np.unique(hub_array, axis=0, return_index=True)

        shroud_composition = find_between(
            channel_text, "NI_BEGIN channel_curve shroud", "NI_END channel_curve shroud"
        )

        if len(shroud_composition) == 0:
            shroud_composition = find_between(
                channel_text,
                "NI_BEGIN\tchannel_curve shroud",
                "NI_END\tchannel_curve shroud",
            )

            if len(shroud_composition) == 0:
                raise ValueError("Cannot define hub curve composition")

        shroud_composition = shroud_composition[0]

        shroud_curves = [
            f"curve_{item}"
            for item in [
                int(match.group(1))
                for match in re.finditer(r"curve_(\d+)", shroud_composition)
            ]
        ]

        if len(shroud_curves) == 0:
            shroud_curves = [
                f"shroud_crv_{item}"
                for item in [
                    int(match.group(1))
                    for match in re.finditer(r"shroud_crv_(\d+)", shroud_composition)
                ]
            ]

        seen = {}
        shroud_curves = [seen.setdefault(x, x) for x in shroud_curves if x not in seen]
        shroud_arrays = [
            ArrayWithName.get_curve_by_name(extracted_curves, [item]).array
            for item in shroud_curves
        ]
        shroud_arrays = self.__fill_straight_curves(shroud_arrays)
        shroud_array = np.concatenate(shroud_arrays)

        # shroud_array, _ = np.unique(shroud_array, axis=0, return_index=True)

        hub_array = np.concatenate(
            [
                np.zeros(shape=(hub_array.shape[0], 1)),
                np.flip(hub_array, axis=-1),
                np.zeros(shape=(hub_array.shape[0], 1)),
            ],
            axis=-1,
        )

        shroud_array = np.concatenate(
            [
                np.zeros(shape=(shroud_array.shape[0], 1)),
                np.flip(shroud_array, axis=-1),
                np.ones(shape=(shroud_array.shape[0], 1)),
            ],
            axis=-1,
        )

        return np.expand_dims(hub_array, axis=0), np.expand_dims(shroud_array, axis=0)

    def __fill_straight_curves(self, curves_list) -> list:
        new_curves = []
        for i, curve in enumerate(curves_list):
            if curve.shape[0] == 2:
                ref = np.array([0, 1])
                x = np.linspace(0, 1, num=self.__straight_lines_filling_points)
                r = np.interp(x, ref, curve[:, 0]).reshape(-1, 1)
                z = np.interp(x, ref, curve[:, 1]).reshape(-1, 1)

                new_curve = np.concatenate([r, z], axis=1)
            else:
                new_curve = curve

            if i != len(curves_list) - 1:
                new_curve = np.delete(new_curve, axis=0, obj=-1)

            new_curves.append(new_curve)
        return new_curves


if __name__ == "__main__":
    # target = "./tests/data/cc_autoblade.geomTurbo"
    target = "./tests/data/rotor_db.geomTurbo"
    diff_active = False
    sp_active = False

    start = time.time()
    parser = GeomTurboParser(
        target_path=target, diffuser_active=diff_active, splitter_active=sp_active
    )
    new_execution_time = time.time() - start
    print("Geomturbo reading execution time: ", new_execution_time)

    # parser.plot_channel(True)
    start = time.time()
    old_parser = numecaParser(filename=target)
    old_execution_time = time.time() - start
    print("OLD Geomturbo reading execution time: ", old_execution_time)

    print("Speedup: %.4f" % (old_execution_time / new_execution_time))

    old_mb = old_parser.exportNpyArray(0, 0)
    old_sp = old_parser.exportNpyArray(0, 1) if sp_active else None
    old_diff = old_parser.exportNpyArray(1, 0) if diff_active else None
    old_hub, old_shroud = old_parser.exportZRNpyArrays()

    parser.plot_channel(show=True)

    assert np.allclose(old_mb, parser.main_blade)
    if sp_active:
        assert np.allclose(old_sp, parser.splitter)
    if diff_active:
        assert np.allclose(old_diff, parser.diffuser)



class numecaEntry:
    def __init__(self, *args):
        self.leadingSpaceString = ""
        self.key = ""
        self.keyTagSpaceString = ""
        self.tag = ""
        self.tagValueSpaceString = ""
        self.value = None
        self.trailingSpaceString = ""
        if len(args) > 0 and isinstance(args[0], numecaEntry):
            self.leadingSpaceString = args[0].leadingSpaceString
            self.key = args[0].key
            self.keyTagSpaceString = args[0].keyTagSpaceString
            self.tag = args[0].tag
            self.tagValueSpaceString = args[0].tagValueSpaceString
            self.value = args[0].value
            self.trailingSpaceString = args[0].trailingSpaceString

    def copy(self, fromObject):
        self.leadingSpaceString = fromObject.leadingSpaceString
        self.key = fromObject.key
        self.keyTagSpaceString = fromObject.keyTagSpaceString
        self.tag = fromObject.tag
        self.tagValueSpaceString = fromObject.tagValueSpaceString
        self.value = fromObject.value
        self.trailingSpaceString = fromObject.trailingSpaceString

    def outputString(self):
        valueString = ""
        if isinstance(self.value, str):
            valueString += self.tagValueSpaceString + self.value
        elif isinstance(self.value, list):
            for idx, token in enumerate(self.value):
                valueString += self.tagValueSpaceString[idx] + token
        return (
            self.leadingSpaceString
            + self.key
            + self.keyTagSpaceString
            + self.tag
            + valueString
            + self.trailingSpaceString
        )

    def __repr__(self):
        valueString = ""
        if isinstance(self.value, str):
            valueString += str(len(self.tagValueSpaceString)) + " " + self.value + "\n"
        elif isinstance(self.value, list):
            for idx, token in enumerate(self.value):
                valueString += (
                    str(len(self.tagValueSpaceString[idx])) + " " + token + "\n"
                )

        return (
            str(len(self.leadingSpaceString))
            + " "
            + self.key
            + "\n"
            + str(len(self.keyTagSpaceString))
            + " "
            + self.tag
            + "\n"
            + valueString
        )

    def __str__(self):
        valueString = ""
        if isinstance(self.value, str):
            valueString += self.value + "\n"
        elif isinstance(self.value, list):
            for idx, token in enumerate(self.value):
                valueString += self.tagValueSpaceString[idx] + token
            valueString += "\n"

        return self.key + "\n" + self.tag + "\n" + valueString
